from .base_kinetics import BaseKinetics
from .total_main_kinetics import TotalMainKinetics
from .butler_volmer import SymmetricButlerVolmer, AsymmetricButlerVolmer
from .linear import Linear
from .marcus import Marcus, MarcusHushChidsey
from .tafel import ForwardTafel  # , BackwardTafel
from .no_reaction import NoReaction
from .msmr_butler_volmer import MSMRButlerVolmer
from .diffusion_limited import DiffusionLimited
from .inverse_kinetics.inverse_butler_volmer import (
    InverseButlerVolmer,
    CurrentForInverseButlerVolmer,
    CurrentForInverseButlerVolmerLithiumMetal,
)

__all__ = ['base_kinetics', 'butler_volmer', 'diffusion_limited',
           'inverse_kinetics', 'linear', 'marcus', 'msmr_butler_volmer',
           'no_reaction', 'tafel', 'total_main_kinetics']



#
# A model to calculate electrode-specific SOH, adapted to a half-cell
#
import pybamm

from .util import _get_lithiation_delithiation, check_if_composite


class ElectrodeSOHHalfCell(pybamm.BaseModel):
    """
    Model to calculate electrode-specific SOH for a half-cell, adapted from
    :footcite:t:`Mohtat2019`.
    This model is mainly for internal use, to calculate summary variables in a
    simulation.

    .. math::
        V_{max} = U_w(x_{100}),
    .. math::
        V_{min} = U_w(x_{0}),
    .. math::
        x_0 = x_{100} - \\frac{C}{C_w}.

    Subscript w indicates working electrode and subscript c indicates counter electrode.

    """

    def __init__(self, name="ElectrodeSOH model", direction=None, options=None):
        pybamm.citations.register("Mohtat2019")
        super().__init__(name)
        options = options or {"working electrode": "positive"}
        param = pybamm.LithiumIonParameters(options)
        is_composite = check_if_composite(options, "positive")

        # Primary phase
        x_100 = pybamm.Variable("x_100", bounds=(0, 1))
        x_0 = pybamm.Variable("x_0", bounds=(0, 1))
        Q_w = pybamm.InputParameter("Q_w")
        T_ref = param.T_ref
        U_w = param.p.prim.U
        lith_delith_primary = _get_lithiation_delithiation(
            direction, "positive", options, phase="primary"
        )

        # Secondary phase
        Q_2 = pybamm.Scalar(0)
        if is_composite:
            x_100_2 = pybamm.Variable("x_100_2", bounds=(0, 1))
            x_0_2 = pybamm.Variable("x_0_2", bounds=(0, 1))
            Q_w_2 = pybamm.InputParameter("Q_w_2")
            U_w_2 = param.p.sec.U
            Q_2 = Q_w_2 * (x_0_2 - x_100_2)
            lith_delith_secondary = _get_lithiation_delithiation(
                direction, "positive", options, phase="secondary"
            )

        Q_1 = Q_w * (x_0 - x_100)
        Q = Q_1 + Q_2

        V_max = param.ocp_soc_100
        V_min = param.ocp_soc_0

        self.algebraic = {
            x_100: U_w(x_100, T_ref, lith_delith_primary) - V_max,
            x_0: U_w(x_0, T_ref, lith_delith_primary) - V_min,
        }
        self.initial_conditions = {x_100: 0.8, x_0: 0.2}
        if is_composite:
            self.algebraic[x_100_2] = (
                U_w_2(x_100_2, T_ref, lith_delith_secondary) - V_max
            )
            self.algebraic[x_0_2] = U_w_2(x_0_2, T_ref, lith_delith_secondary) - V_min
            self.initial_conditions[x_100_2] = 0.8
            self.initial_conditions[x_0_2] = 0.2

        self.variables = {
            "x_100": x_100,
            "x_0": x_0,
            "Q": Q,
            "Uw(x_100)": U_w(x_100, T_ref, lith_delith_primary),
            "Uw(x_0)": U_w(x_0, T_ref, lith_delith_primary),
            "Q_w": Q_w,
        }
        if is_composite:
            self.variables.update(
                {
                    "x_100_2": x_100_2,
                    "x_0_2": x_0_2,
                    "Q_w_2": Q_w_2,
                    "Uw(x_100_2)": U_w_2(x_100_2, T_ref, lith_delith_secondary),
                    "Uw(x_0_2)": U_w_2(x_0_2, T_ref, lith_delith_secondary),
                }
            )

    @property
    def default_solver(self):
        # Use AlgebraicSolver as CasadiAlgebraicSolver gives unnecessary warnings
        return pybamm.AlgebraicSolver(method="lsq__trf", tol=1e-7)


def get_initial_stoichiometry_half_cell(
    initial_value,
    parameter_values,
    param=None,
    options=None,
    tol=1e-6,
    inputs=None,
    direction=None,
    **kwargs,
):
    """
    Calculate initial stoichiometry to start off the simulation at a particular
    state of charge, given voltage limits, open-circuit potential, etc defined by
    parameter_values

    Parameters
    ----------
    initial_value : float
        Target initial value.
        If integer, interpreted as SOC, must be between 0 and 1.
        If string e.g. "4 V", interpreted as voltage,
        must be between V_min and V_max.
    parameter_values : pybamm.ParameterValues
        The parameter values to use in the calculation
    param : :class:`pybamm.LithiumIonParameters`, optional
        The symbolic parameter set to use for the simulation.
        If not provided, the default parameter set will be used.
    tol : float, optional
        The tolerance for the solver used to compute the initial stoichiometries.
        Default is 1e-6.
    inputs : dict, optional
        A dictionary of input parameters passed to the model.

    Returns
    -------
    x
        The initial stoichiometry that give the desired initial state of charge
    """
    param = pybamm.LithiumIonParameters(options)
    x_dict = get_min_max_stoichiometries(
        parameter_values, inputs=inputs, direction=direction, options=options
    )
    x_0, x_100 = x_dict["x_0"], x_dict["x_100"]
    is_composite = check_if_composite(options, "positive")
    lith_delith_primary = _get_lithiation_delithiation(
        direction, "positive", options, phase="primary"
    )
    if is_composite:
        x_0_2, x_100_2 = x_dict["x_0_2"], x_dict["x_100_2"]
        lith_delith_secondary = _get_lithiation_delithiation(
            direction, "positive", options, phase="secondary"
        )

    if isinstance(initial_value, str) and initial_value.endswith("V"):
        V_init = float(initial_value[:-1])
        V_min = parameter_values.evaluate(param.voltage_low_cut, inputs=inputs)
        V_max = parameter_values.evaluate(param.voltage_high_cut, inputs=inputs)

        if not V_min - tol <= V_init <= V_max + tol:
            raise ValueError(
                f"Initial voltage {V_init}V is outside the voltage limits "
                f"({V_min}, {V_max})"
            )
        # Solve simple model for initial soc based on target voltage
        model = pybamm.BaseModel()
        x = pybamm.Variable("x")
        Up = param.p.prim.U
        T_init = param.T_init
        model.variables["x"] = x
        model.algebraic[x] = Up(x, T_init, lith_delith_primary) - V_init
        # initial guess for x linearly interpolates between 0 and 1
        # based on V linearly interpolating between V_max and V_min
        soc_initial_guess = (V_init - V_min) / (V_max - V_min)
        model.initial_conditions[x] = 1 - soc_initial_guess
        if is_composite:
            Up_2 = param.p.sec.U
            x_2 = pybamm.Variable("x_2")
            model.algebraic[x_2] = Up_2(x_2, T_init, lith_delith_secondary) - V_init
            model.variables["x_2"] = x_2
            model.initial_conditions[x_2] = 1 - soc_initial_guess

        parameter_values.process_model(model)
        sol = pybamm.AlgebraicSolver("lsq__trf", tol=tol).solve(
            model, [0], inputs=inputs
        )
        x = sol["x"].data[0]
        if is_composite:
            x_2 = sol["x_2"].data[0]
    elif isinstance(initial_value, int | float):
        if not 0 <= initial_value <= 1:
            raise ValueError("Initial SOC should be between 0 and 1")
        if not is_composite:
            x = x_0 + initial_value * (x_100 - x_0)
        else:
            model = pybamm.BaseModel()
            x = pybamm.Variable("x")
            x_2 = pybamm.Variable("x_2")
            U_p = param.p.prim.U
            U_p_2 = param.p.sec.U
            # here we use T_ref and SOC 0 and 1 are defined using the reference state
            T_ref = param.T_ref
            model.algebraic[x] = U_p(x, T_ref, lith_delith_primary) - U_p_2(
                x_2, T_ref, lith_delith_secondary
            )
            model.initial_conditions[x] = x_0 + initial_value * (x_100 - x_0)
            model.initial_conditions[x_2] = x_0_2 + initial_value * (x_100_2 - x_0_2)
            Q_w = parameter_values.evaluate(param.p.prim.Q_init, inputs=inputs)
            Q_w_2 = parameter_values.evaluate(param.p.sec.Q_init, inputs=inputs)
            Q_min = x_100 * Q_w + x_100_2 * Q_w_2
            Q_max = x_0 * Q_w + x_0_2 * Q_w_2
            Q_now = Q_w * x + Q_w_2 * x_2
            soc = (Q_now - Q_min) / (Q_max - Q_min)
            model.algebraic[x_2] = soc - initial_value
            model.variables["x"] = x
            model.variables["x_2"] = x_2
            parameter_values.process_model(model)
            sol = pybamm.AlgebraicSolver(tol=tol).solve(model, [0], inputs=inputs)
            x = sol["x"].data[0]
            x_2 = sol["x_2"].data[0]
    else:
        raise ValueError(
            "Initial value must be a float between 0 and 1, or a string ending in 'V'"
        )
    ret_dict = {"x": x}
    if is_composite:
        ret_dict["x_2"] = x_2
    return ret_dict


def get_min_max_stoichiometries(
    parameter_values, options=None, inputs=None, direction=None
):
    """
    Get the minimum and maximum stoichiometries from the parameter values

    Parameters
    ----------
    parameter_values : pybamm.ParameterValues
        The parameter values to use in the calculation
    options : dict, optional
        A dictionary of options to be passed to the parameters, see
        :class:`pybamm.BatteryModelOptions`.
        If None, the default is used: {"working electrode": "positive"}
    """
    inputs = inputs or {}
    if options is None:
        options = {"working electrode": "positive"}
    esoh_model = pybamm.lithium_ion.ElectrodeSOHHalfCell(
        "ElectrodeSOH", direction=direction, options=options
    )
    param = pybamm.LithiumIonParameters(options)
    is_composite = check_if_composite(options, "positive")
    if is_composite:
        Q_w = parameter_values.evaluate(param.p.prim.Q_init, inputs=inputs)
        Q_w_2 = parameter_values.evaluate(param.p.sec.Q_init, inputs=inputs)
        Q_inputs = {"Q_w": Q_w, "Q_w_2": Q_w_2}
    else:
        Q_w = parameter_values.evaluate(param.p.prim.Q_init, inputs=inputs)
        Q_inputs = {"Q_w": Q_w}
    # Add Q_w to input parameters
    all_inputs = {**inputs, **Q_inputs}
    esoh_sim = pybamm.Simulation(esoh_model, parameter_values=parameter_values)
    esoh_sol = esoh_sim.solve([0], inputs=all_inputs)
    x_0, x_100 = esoh_sol["x_0"].data[0], esoh_sol["x_100"].data[0]
    if is_composite:
        x_0_2, x_100_2 = esoh_sol["x_0_2"].data[0], esoh_sol["x_100_2"].data[0]
        ret_dict = {"x_0": x_0, "x_100": x_100, "x_0_2": x_0_2, "x_100_2": x_100_2}
    else:
        ret_dict = {"x_0": x_0, "x_100": x_100}
    return ret_dict



from pybamm.version import __version__

# Utility classes and methods
from .util import root_dir
from .util import Timer, TimerTime, FuzzyDict
from .util import (
    root_dir,
    load,
    is_constant_and_can_evaluate,
)
from .util import (
    get_parameters_filepath,
    has_jax,
    import_optional_dependency,
)
from .logger import logger, set_logging_level, get_new_logger
from .settings import settings
from .citations import Citations, citations, print_citations
from . import config

# Classes for the Expression Tree
from .expression_tree.symbol import *
from .expression_tree.binary_operators import *
from .expression_tree.concatenations import *
from .expression_tree.array import Array, linspace, meshgrid
from .expression_tree.matrix import Matrix
from .expression_tree.unary_operators import *
from .expression_tree.averages import *
from .expression_tree.averages import _BaseAverage
from .expression_tree.broadcasts import *
from .expression_tree.functions import *
from .expression_tree.interpolant import Interpolant
from .expression_tree.discrete_time_sum import *
from .expression_tree.input_parameter import InputParameter
from .expression_tree.parameter import Parameter, FunctionParameter
from .expression_tree.scalar import Scalar, Constant
from .expression_tree.variable import *
from .expression_tree.coupled_variable import *
from .expression_tree.independent_variable import *
from .expression_tree.independent_variable import t
from .expression_tree.vector import Vector
from .expression_tree.vector_field import VectorField
from .expression_tree.state_vector import StateVectorBase, StateVector, StateVectorDot

from .expression_tree.exceptions import *

# Operations
from .expression_tree.operations.evaluate_python import (
    find_symbols,
    id_to_python_variable,
    to_python,
    EvaluatorPython,
)

from .expression_tree.operations.evaluate_python import EvaluatorJax
from .expression_tree.operations.evaluate_python import JaxCooMatrix

from .expression_tree.operations.jacobian import Jacobian
from .expression_tree.operations.convert_to_casadi import CasadiConverter
from .expression_tree.operations.unpack_symbols import SymbolUnpacker
from .expression_tree.operations.serialise import Serialise,ExpressionFunctionParameter

# Model classes
from .models.base_model import BaseModel
from .models.event import Event
from .models.event import EventType

# Battery models
from .models.full_battery_models.base_battery_model import (
    BaseBatteryModel,
    BatteryModelOptions,
)
from .models.full_battery_models import lead_acid
from .models.full_battery_models import lithium_ion
from .models.full_battery_models import equivalent_circuit
from .models.full_battery_models import sodium_ion

# Submodel classes
from .models.submodels.base_submodel import BaseSubModel

from .models.submodels import (
    active_material,
    convection,
    current_collector,
    electrolyte_conductivity,
    electrolyte_diffusion,
    electrode,
    external_circuit,
    interface,
    oxygen_diffusion,
    particle,
    porosity,
    thermal,
    transport_efficiency,
    particle_mechanics,
    equivalent_circuit_elements,
)
from .models.submodels.interface import kinetics
from .models.submodels.interface import sei
from .models.submodels.interface import lithium_plating
from .models.submodels.interface import interface_utilisation
from .models.submodels.interface import open_circuit_potential

# Geometry
from .geometry.geometry import Geometry
from .geometry.battery_geometry import battery_geometry

from .expression_tree.independent_variable import KNOWN_COORD_SYS
from .geometry import standard_spatial_vars

# Parameter classes and methods
from .parameters.parameter_values import ParameterValues, scalarize_dict, arrayize_dict
from .parameters import constants
from .parameters.geometric_parameters import geometric_parameters, GeometricParameters
from .parameters.electrical_parameters import (
    electrical_parameters,
    ElectricalParameters,
)
from .parameters.thermal_parameters import thermal_parameters, ThermalParameters
from .parameters.lithium_ion_parameters import LithiumIonParameters
from .parameters.lead_acid_parameters import LeadAcidParameters
from .parameters.ecm_parameters import EcmParameters
from .parameters.size_distribution_parameters import *

# Mesh and Discretisation classes
from .discretisations.discretisation import Discretisation
from .discretisations.discretisation import has_bc_of_form
from .meshes.meshes import Mesh, SubMesh, MeshGenerator
from .meshes.zero_dimensional_submesh import SubMesh0D
from .meshes.one_dimensional_submeshes import (
    SubMesh1D,
    Uniform1DSubMesh,
    Exponential1DSubMesh,
    Chebyshev1DSubMesh,
    UserSupplied1DSubMesh,
    SpectralVolume1DSubMesh,
    SymbolicUniform1DSubMesh,
)
from .meshes.two_dimensional_submeshes import (
    SubMesh2D,
    Uniform2DSubMesh,
)
from .meshes.scikit_fem_submeshes import (
    ScikitSubMesh2D,
    ScikitUniform2DSubMesh,
    ScikitExponential2DSubMesh,
    ScikitChebyshev2DSubMesh,
    UserSupplied2DSubMesh,
)

from .meshes.scikit_fem_submeshes_3d import (
    ScikitFemSubMesh3D,
    ScikitFemGenerator3D,
    UserSuppliedSubmesh3D,
)

# Serialisation
from .models.base_model import load_model

# Spatial Methods
from .spatial_methods.spatial_method import SpatialMethod
from .spatial_methods.zero_dimensional_method import ZeroDimensionalSpatialMethod
from .spatial_methods.finite_volume import FiniteVolume
from .spatial_methods.finite_volume_2d import FiniteVolume2D
from .spatial_methods.spectral_volume import SpectralVolume
from .spatial_methods.scikit_finite_element import ScikitFiniteElement
from .spatial_methods.scikit_finite_element_3d import ScikitFiniteElement3D

# Solver classes
from .solvers.solution import Solution, EmptySolution, make_cycle_solution
from .solvers.processed_variable_time_integral import ProcessedVariableTimeIntegral
from .solvers.processed_variable import ProcessedVariable, ProcessedVariable2DFVM, process_variable
from .solvers.processed_variable_computed import ProcessedVariableComputed
from .solvers.processed_variable import ProcessedVariableUnstructured
from .solvers.summary_variable import SummaryVariables
from .solvers.base_solver import BaseSolver
from .solvers.dummy_solver import DummySolver
from .solvers.algebraic_solver import AlgebraicSolver
from .solvers.casadi_solver import CasadiSolver
from .solvers.casadi_algebraic_solver import CasadiAlgebraicSolver
from .solvers.scipy_solver import ScipySolver

from .solvers.jax_solver import JaxSolver
from .solvers.jax_bdf_solver import jax_bdf_integrate

from .solvers.idaklu_jax import IDAKLUJax
from .solvers.idaklu_solver import IDAKLUSolver

# Experiments
from .experiment.experiment import Experiment
from . import experiment
from .experiment import step

# Plotting
from .plotting.quick_plot import QuickPlot, close_plots, QuickPlotAxes
from .plotting.plot import plot
from .plotting.plot2D import plot2D
from .plotting.plot_voltage_components import plot_voltage_components
from .plotting.plot_thermal_components import plot_thermal_components
from .plotting.plot_summary_variables import plot_summary_variables
from .plotting.dynamic_plot import dynamic_plot
from .plotting.plot_3d_cross_section import plot_3d_cross_section
from .plotting.plot_3d_heatmap import plot_3d_heatmap

# Simulation
from .simulation import Simulation, load_sim, is_notebook

# Batch Study
from .batch_study import BatchStudy

# Callbacks, telemetry, config
from . import callbacks, telemetry, config

# Pybamm Data manager using pooch
from .pybamm_data import DataLoader

from .dispatch import parameter_sets, Model

# Fix Casadi import
import os
import pathlib
import sysconfig

os.environ["CASADIPATH"] = str(pathlib.Path(sysconfig.get_path("purelib")) / "casadi")

__all__ = [
    "batch_study",
    "callbacks",
    "citations",
    "config",
    "discretisations",
    "doc_utils",
    "experiment",
    "expression_tree",
    "geometry",
    "input",
    "logger",
    "meshes",
    "models",
    "parameters",
    "plotting",
    "settings",
    "simulation",
    "solvers",
    "spatial_methods",
    "telemetry",
    "type_definitions",
    "util",
    "version",
    "pybamm_data",
    "dispatch",
]

config.generate()



from __future__ import annotations

from typing import Any, TypeAlias

import numpy as np
import numpy.typing as npt

import pybamm

# numbers.Number should not be used for type hints
Numeric: TypeAlias = int | float | np.number

# expression tree
ChildValue: TypeAlias = float | npt.NDArray[Any]
ChildSymbol: TypeAlias = float | npt.NDArray[Any] | pybamm.Symbol

DomainType: TypeAlias = list[str] | str | None
AuxiliaryDomainType: TypeAlias = dict[str, str] | None
DomainsType: TypeAlias = dict[str, list[str] | str] | None



import importlib.metadata
import textwrap
from collections.abc import Callable, Mapping


class EntryPoint(Mapping):
    """
    A mapping interface for accessing PyBaMM models and parameter sets through entry points.

    This class provides a unified way to load, and instantiate PyBaMM models
    and parameter sets that have been registered as entry points.

    Access via :py:data:`pybamm.parameter_sets` for parameter sets - provides access
    to all registered battery parameter sets (e.g., 'Chen2020', 'Ai2020') that can
    be used to parameterise battery models.

    Access via :py:data:`pybamm.Model` for models - provides access to all registered
    battery models (e.g., 'SPM', 'DFN').

    .. attention::

        This feature is currently experimental.

    Examples
    --------
    Listing available parameter sets:
        >>> import pybamm
        >>> list(pybamm.parameter_sets)  # doctest: +ELLIPSIS
        ['Ai2020', 'Chayambuka2022', ...]

    Listing available models:
        >>> import pybamm
        >>> list(pybamm.dispatch.models) # doctest: +ELLIPSIS
        ['DFN', 'MPM', ...]

    Get the docstring for a parameter set or model:
        >>> print(pybamm.parameter_sets.get_docstring("Ai2020"))  # doctest: +ELLIPSIS
        <BLANKLINE>
        Parameters for the Enertech cell (Ai2020), from the papers :footcite:t:`Ai2019`...

        >>> print(pybamm.dispatch.models.get_docstring("SPM"))  # doctest: +ELLIPSIS
        <BLANKLINE>
        Single Particle Model (SPM) of a lithium-ion battery...
    """

    _instances = 0

    def __init__(self, group):
        """Dict of entry points for parameter sets or models to lazily load as entry points"""
        if not hasattr(
            self, "initialized"
        ):  # Ensure __init__ is called once per instance
            self.initialized = True
            EntryPoint._instances += 1
            self._all_entries = dict()
            self.group = group
            for entry_point in self.get_entries(self.group):
                self._all_entries[entry_point.name] = entry_point

    @staticmethod
    def get_entries(group_name):
        """Wrapper for the importlib version logic"""
        return importlib.metadata.entry_points(group=group_name)

    def __new__(cls, group):
        """Ensure only two instances of entry points exist, one for parameter sets and the other for models"""
        if EntryPoint._instances < 2:
            cls.instance = super().__new__(cls)
        return cls.instance

    def _get_class(self, key) -> Callable:
        """Return the class without instantiating it"""
        return self._load_entry_point(key)

    def __getitem__(self, key) -> dict:
        return self._load_entry_point(key)()

    def _load_entry_point(self, key) -> Callable:
        """Check that ``key`` is a registered ``parameter_sets`` or ``models` ,
        and return the entry point for the parameter set/model, loading it needed."""
        if key not in self._all_entries:
            raise KeyError(f"Unknown parameter set or model: {key}")
        entry_point = self._all_entries[key]
        try:
            entry_point = self._all_entries[key] = entry_point.load()
        except AttributeError:
            # Parameter sets cannot be loaded so returning the default entry_point
            pass
        return entry_point

    def __iter__(self):
        return self._all_entries.__iter__()

    def __len__(self) -> int:
        return len(self._all_entries)

    def get_docstring(self, key):
        """Return the docstring for the ``key`` parameter set or model"""
        return textwrap.dedent(self._load_entry_point(key).__doc__)

    def __getattribute__(self, name):
        return super().__getattribute__(name)


#: Singleton Instance of :class:EntryPoint initialised with pybamm_parameter_sets"""
parameter_sets = EntryPoint(group="pybamm_parameter_sets")

#: Singleton Instance of :class:EntryPoint initialised with pybamm_models"""
models = EntryPoint(group="pybamm_models")


def Model(model: str, *args, **kwargs):
    """
    Returns the loaded model object
    Note: This feature is in its experimental phase.

    Parameters
    ----------
    model : str
        The model name or author name of the model mentioned at the model entry point.
    *args
        Additional positional arguments to pass to the model constructor.
    **kwargs
        Additional keyword arguments to pass to the model constructor.

    Returns
    -------
    pybamm.model
        Model object of the initialised model.

    Examples
    --------
    Listing available models:
        >>> import pybamm
        >>> list(pybamm.dispatch.models) # doctest: +ELLIPSIS
        ['DFN', 'MPM', ...]
        >>> pybamm.Model('SPM') # doctest: +SKIP
        <pybamm.models.full_battery_models.lithium_ion.spm.SPM object>
    """
    model_class = models._get_class(model)

    return model_class(*args, **kwargs)



import re

import pybamm

from .dfn import DFN


class MSMR(DFN):
    def __init__(self, options=None, name="MSMR", build=True):
        # Necessary/default options
        options = options or {}
        if "number of MSMR reactions" not in options:
            raise pybamm.OptionError(
                "number of MSMR reactions must be specified for MSMR"
            )
        if (
            "open-circuit potential" in options
            and options["open-circuit potential"] != "MSMR"
        ):
            raise pybamm.OptionError(
                "'open-circuit potential' must be 'MSMR' for MSMR not '{}'".format(
                    options["open-circuit potential"]
                )
            )
        elif "particle" in options and options["particle"] != "MSMR":
            raise pybamm.OptionError(
                "'particle' must be 'MSMR' for MSMR not '{}'".format(
                    options["particle"]
                )
            )
        elif (
            "intercalation kinetics" in options
            and options["intercalation kinetics"] != "MSMR"
        ):
            raise pybamm.OptionError(
                "'intercalation kinetics' must be 'MSMR' for MSMR not '{}'".format(
                    options["intercalation kinetics"]
                )
            )
        else:
            options.update(
                {
                    "open-circuit potential": "MSMR",
                    "particle": "MSMR",
                    "intercalation kinetics": "MSMR",
                }
            )
        super().__init__(options=options, name=name)

    @property
    def default_parameter_values(self):
        return pybamm.ParameterValues("MSMR_Example")


# Replace the deprecated MSMR parameter names with the new ones
# matches e.g. "X_p_3", "X_n_l_300", "Q_n_d_0", "U0_p_42", "a_n_5", "j0_ref_p_d_12", etc.
_VALID_NAME_RE = re.compile(
    r"^(?P<base>X|Q|w|U0|a|j0_ref)"  # base (now includes a and j0_ref)
    r"_(?P<elec>n|p)"  # electrode
    r"(?:_(?P<qual>[ld]))?"  # optional qualifier
    r"_(?P<idx>[0-9]+)$"  # non-negative integer index
)


def is_deprecated_msmr_name(key: str) -> bool:
    """
    Return True if `key` follows the (legacy) MSMR naming convention:
      BASE ∈ {X, Q, w, U0, a, j0_ref}
      electrode ∈ {n, p}
      optional qualifier ∈ {l, d}
      index ∈ non-negative integer
    """
    return bool(_VALID_NAME_RE.fullmatch(key))


_BASE_DESC = {
    "X": "host site occupancy fraction",
    "Q": "host site occupancy capacity",
    "w": "host site ideality factor",
    "U0": "host site standard potential",
    "a": "host site charge transfer coefficient",
    "j0_ref": "host site reference exchange-current density",
}

# only Q, U0, and j0_ref have units
_UNITS = {
    "Q": " [A.h]",
    "U0": " [V]",
    "j0_ref": " [A.m-2]",
}

_ELECTRODE = {
    "n": "Negative",
    "p": "Positive",
}

_QUALIFIER = {
    "l": "lithiation",
    "d": "delithiation",
}


def replace_deprecated_msmr_name(key: str) -> str:
    """
    Convert e.g. "X_n_d_3" →
        "Negative electrode host site occupancy fraction (delithiation) (3)"
    and likewise for a/U0/Q/j0_ref.
    """
    m = _VALID_NAME_RE.fullmatch(key)
    if not m:
        raise ValueError(f"Invalid MSMR name: {key!r}")

    base = m.group("base")
    elec = m.group("elec")
    qual = m.group("qual")
    idx = m.group("idx")

    # start constructing the description
    desc = f"{_ELECTRODE[elec]} electrode {_BASE_DESC[base]}"
    if qual:
        desc += f" ({_QUALIFIER[qual]})"
    desc += f" ({idx})"

    # tack on units if this base has them
    if base in _UNITS:
        desc += _UNITS[base]

    return desc



#
# Marcus and Asymptotic Marcus-Hush-Chidsey classes
#

import numpy as np

import pybamm

from .base_kinetics import BaseKinetics


class Marcus(BaseKinetics):
    """
    Submodel which implements Marcus kinetics.

    Parameters
    ----------
    param : parameter class
        model parameters
    domain : str
        The domain to implement the model, either: 'Negative' or 'Positive'.
    reaction : str
        The name of the reaction being implemented
    options: dict
        A dictionary of options to be passed to the model.
        See :class:`pybamm.BaseBatteryModel`
    phase : str, optional
        Phase of the particle (default is "primary")
    """

    def __init__(self, param, domain, reaction, options, phase="primary"):
        super().__init__(param, domain, reaction, options, phase)
        pybamm.citations.register("Sripad2020")

    def _get_kinetics(self, j0, ne, eta_r, T, u):
        RT = self.param.R * T
        Feta_RT = self.param.F * eta_r / RT
        mhc_lambda = self.phase_param.mhc_lambda

        exp_arg_ox = -((mhc_lambda + Feta_RT) ** 2) / (4 * mhc_lambda * RT)
        exp_arg_red = -((mhc_lambda - Feta_RT) ** 2) / (4 * mhc_lambda * RT)
        return u * j0 * (pybamm.exp(exp_arg_ox) - pybamm.exp(exp_arg_red))


class MarcusHushChidsey(BaseKinetics):
    """
    Submodel which implements asymptotic Marcus-Hush-Chidsey kinetics, as derived in
    :footcite:t:`Sripad2020`

    Parameters
    ----------
    param : parameter class
        model parameters
    domain : str
        The domain to implement the model, either: 'Negative' or 'Positive'.
    reaction : str
        The name of the reaction being implemented
    options: dict
        A dictionary of options to be passed to the model.
        See :class:`pybamm.BaseBatteryModel`
    phase : str, optional
        Phase of the particle (default is "primary")
    """

    def __init__(self, param, domain, reaction, options, phase="primary"):
        super().__init__(param, domain, reaction, options, phase)
        pybamm.citations.register("Sripad2020")

    def _get_kinetics(self, j0, ne, eta_r, T, u):
        mhc_lambda = self.phase_param.mhc_lambda

        F_RT = self.param.F / (self.param.R * T)
        Feta_RT = F_RT * eta_r
        lambda_T = F_RT * mhc_lambda
        a = 1 + pybamm.sqrt(lambda_T)

        arg = (lambda_T - pybamm.sqrt(a + Feta_RT**2)) / (2 * pybamm.sqrt(lambda_T))
        pref = pybamm.sqrt(np.pi * lambda_T) * pybamm.tanh(Feta_RT / 2)
        return j0 * u * pref * pybamm.erfc(arg)



#
# Class for quick plotting of variables from models
#
from collections import defaultdict

import numpy as np

import pybamm
from pybamm.util import import_optional_dependency


class LoopList(list):
    """A list which loops over itself when accessing an
    index so that it never runs out"""

    def __getitem__(self, i):
        # implement looping by calling "(i) modulo (length of list)"
        return super().__getitem__(i % len(self))


def ax_min(data):
    """Calculate appropriate minimum axis value for plotting"""
    data_min = np.nanmin(data)
    data_max = np.nanmax(data)
    return data_max - 1.05 * (data_max - data_min)


def ax_max(data):
    """Calculate appropriate maximum axis value for plotting"""
    data_min = np.nanmin(data)
    data_max = np.nanmax(data)
    return data_min + 1.05 * (data_max - data_min)


def split_long_string(title, max_words=None):
    """Get title in a nice format"""
    max_words = max_words or pybamm.settings.max_words_in_line
    words = title.split()
    # Don't split if fits on one line, don't split just for units
    if len(words) <= max_words or words[max_words].startswith("["):
        return title
    else:
        first_line = (" ").join(words[:max_words])
        second_line = (" ").join(words[max_words:])
        return first_line + "\n" + second_line


def close_plots():
    """Close all open figures"""
    plt = import_optional_dependency("matplotlib.pyplot")

    plt.close("all")


class QuickPlot:
    """
    Generates a quick plot of a subset of key outputs of the model so that the model
    outputs can be easily assessed.

    Parameters
    ----------
    solutions: (iter of) :class:`pybamm.Solution` or :class:`pybamm.Simulation`
        The numerical solution(s) for the model(s), or the simulation object(s)
        containing the solution(s).
    output_variables : list of str, optional
        List of variables to plot
    labels : list of str, optional
        Labels for the different models. Defaults to model names
    colors : list of str, optional
        The colors to loop over when plotting. Defaults to None, in which case the
        default color loop defined by matplotlib style sheet or rcParams is used.
    linestyles : list of str, optional
        The linestyles to loop over when plotting. Defaults to ["-", ":", "--", "-."]
    shading : str, optional
        The shading to use for 2D plots. Defaults to "auto".
    figsize : tuple of floats, optional
        The size of the figure to make
    n_rows : int, optional
        The number of rows to use. If None (default), floor(n // sqrt(n)) is used where
        n = len(output_variables) so that the plot is as square as possible
    time_unit : str, optional
        Format for the time output ("hours", "minutes", or "seconds")
    spatial_unit : str, optional
        Format for the spatial axes ("m", "mm", or "um")
    variable_limits : str or dict of str, optional
        How to set the axis limits (for 0D or 1D variables) or colorbar limits (for 2D
        variables). Options are:
    n_t_linear: int, optional
        The number of linearly spaced time points added to the t axis for each sub-solution.
        Note: this is only used if the solution has hermite interpolation enabled.

        - "fixed" (default): keep all axes fixes so that all data is visible
        - "tight": make axes tight to plot at each time
        - dictionary: fine-grain control for each variable, can be either "fixed" or \
        "tight" or a specific tuple (lower, upper).

    """

    def __init__(
        self,
        solutions,
        output_variables=None,
        labels=None,
        colors=None,
        linestyles=None,
        shading="auto",
        figsize=None,
        n_rows=None,
        time_unit=None,
        spatial_unit="um",
        variable_limits="fixed",
        n_t_linear=100,
    ):
        solutions = self.preprocess_solutions(solutions)

        models = [solution.all_models[0] for solution in solutions]

        # Set labels
        if labels is None:
            self.labels = [model.name for model in models]
        else:
            if len(labels) != len(models):
                raise ValueError(
                    f"labels '{labels}' have different length to models '{[model.name for model in models]}'"
                )
            self.labels = labels

        # Set colors, linestyles, figsize, axis limits
        # call LoopList to make sure list index never runs out
        if colors is None:
            self.colors = LoopList(colors or ["r", "b", "k", "g", "m", "c"])
        else:
            self.colors = LoopList(colors)
        self.linestyles = LoopList(linestyles or ["-", ":", "--", "-."])
        self.shading = shading

        # Default output variables for lead-acid and lithium-ion
        if output_variables is None:
            output_variables = models[0].default_quick_plot_variables
            # raise error if still None
            if output_variables is None:
                raise ValueError(
                    f"No default output variables provided for {models[0].name}"
                )

        # check variables have been provided after any serialisation
        if any(len(m.variables) == 0 for m in models):
            raise AttributeError("No variables to plot")

        self.n_rows = n_rows or int(
            len(output_variables) // np.sqrt(len(output_variables))
        )
        self.n_cols = int(np.ceil(len(output_variables) / self.n_rows))

        figwidth_default = min(15, 4 * self.n_cols)
        figheight_default = min(8, 1 + 3 * self.n_rows)
        self.figsize = figsize or (figwidth_default, figheight_default)

        # Spatial scales (default to 1 if information not in model)
        if spatial_unit == "m":
            self.spatial_factor = 1
            self.spatial_unit = "m"
        elif spatial_unit == "mm":
            self.spatial_factor = 1e3
            self.spatial_unit = "mm"
        elif spatial_unit == "um":  # micrometers
            self.spatial_factor = 1e6
            self.spatial_unit = r"$\mu$m"
        else:
            raise ValueError(f"spatial unit '{spatial_unit}' not recognized")

        # Time parameters
        self.ts_seconds = [solution.t for solution in solutions]
        min_t = np.min([t[0] for t in self.ts_seconds])
        max_t = np.max([t[-1] for t in self.ts_seconds])

        hermite_interp = all(sol.hermite_interpolation for sol in solutions)

        def t_sample(sol):
            if hermite_interp and n_t_linear > 2:
                # Linearly spaced time points
                t_linspace = np.linspace(sol.t[0], sol.t[-1], n_t_linear + 2)[1:-1]
                t_plot = np.union1d(sol.t, t_linspace)
            else:
                t_plot = sol.t
            return t_plot

        ts_seconds = []
        for sol in solutions:
            # Sample time points for each sub-solution
            t_sol = [t_sample(sub_sol) for sub_sol in sol.sub_solutions]
            ts_seconds.append(np.concatenate(t_sol))
        self.ts_seconds = ts_seconds

        # Set timescale
        if time_unit is None:
            # defaults depend on how long the simulation is
            if max_t >= 3600:
                time_scaling_factor = 3600  # time in hours
                self.time_unit = "h"
            else:
                time_scaling_factor = 1  # time in seconds
                self.time_unit = "s"
        elif time_unit == "seconds":
            time_scaling_factor = 1
            self.time_unit = "s"
        elif time_unit == "minutes":
            time_scaling_factor = 60
            self.time_unit = "min"
        elif time_unit == "hours":
            time_scaling_factor = 3600
            self.time_unit = "h"
        else:
            raise ValueError(f"time unit '{time_unit}' not recognized")
        self.time_scaling_factor = time_scaling_factor
        self.min_t_unscaled = min_t
        self.max_t_unscaled = max_t

        # Prepare dictionary of variables
        # output_variables is a list of strings or lists, e.g.
        # ["var 1", ["variable 2", "var 3"]]
        output_variable_tuples = []
        self.variable_limits = {}
        for variable_list in output_variables:
            # Make sure we always have a list of lists of variables, e.g.
            # [["var 1"], ["variable 2", "var 3"]]
            if isinstance(variable_list, str):
                variable_list = [variable_list]

            # Store the key as a tuple
            variable_tuple = tuple(variable_list)
            output_variable_tuples.append(variable_tuple)

            # axis limits
            if variable_limits in ["fixed", "tight"]:
                self.variable_limits[variable_tuple] = variable_limits
            else:
                # If there is only one variable, extract it
                if len(variable_tuple) == 1:
                    variable = variable_tuple[0]
                else:
                    variable = variable_tuple
                try:
                    self.variable_limits[variable_tuple] = variable_limits[variable]
                except KeyError:
                    # if variable_tuple is not provided, default to "fixed"
                    self.variable_limits[variable_tuple] = "fixed"
                except TypeError as error:
                    raise TypeError(
                        "variable_limits must be 'fixed', 'tight', or a dict"
                    ) from error

        self.set_output_variables(output_variable_tuples, solutions)
        self.reset_axis()

    @staticmethod
    def preprocess_solutions(solutions):
        input_solutions = QuickPlot.check_input_validity(solutions)
        processed_solutions = []
        for sim_or_sol in input_solutions:
            if isinstance(sim_or_sol, pybamm.Simulation):
                # 'sim_or_sol' is actually a 'Simulation' object here, so it has a
                # 'Solution' attribute
                processed_solutions.append(sim_or_sol.solution)
            elif isinstance(sim_or_sol, pybamm.Solution):
                processed_solutions.append(sim_or_sol)
        return processed_solutions

    @staticmethod
    def check_input_validity(input_solutions):
        if not isinstance(input_solutions, pybamm.Solution | pybamm.Simulation | list):
            raise TypeError(
                "Solutions must be 'pybamm.Solution' or 'pybamm.Simulation' or list"
            )
        elif not isinstance(input_solutions, list):
            input_solutions = [input_solutions]
        else:
            if not input_solutions:
                raise TypeError("QuickPlot requires at least 1 solution or simulation.")
        return input_solutions

    def set_output_variables(self, output_variables, solutions):
        # Set up output variables
        self.variables = {}
        self.spatial_variable_dict = {}
        self.first_spatial_variable = {}
        self.second_spatial_variable = {}
        self.x_first_and_y_second = {}
        self.is_y_z = {}

        # Calculate subplot positions based on number of variables supplied
        self.subplot_positions = {}

        for k, variable_tuple in enumerate(output_variables):
            # Prepare list of variables
            variables = [None] * len(solutions)

            # process each variable in variable_list for each model
            for i, solution in enumerate(solutions):
                # variables lists of lists, so variables[i] is a list
                variables[i] = []
                for var in variable_tuple:
                    sol = solution[var]
                    # Check variable isn't all-nan
                    if np.all(np.isnan(sol.entries)):
                        raise ValueError(f"All-NaN variable '{var}' provided")
                    # If ok, add to the list of solutions
                    else:
                        variables[i].append(sol)

            # Make sure variables have the same dimensions and domain
            # just use the first solution to check this
            first_solution = variables[0]
            first_variable = first_solution[0]
            domain = first_variable.domain
            # check all other solutions against the first solution
            for idx, variable in enumerate(first_solution):
                if variable.domain != domain:
                    raise ValueError(
                        "Mismatching variable domains. "
                        f"'{variable_tuple[0]}' has domain '{domain}', but '{variable_tuple[idx]}' has domain '{variable.domain}'"
                    )
                self.spatial_variable_dict[variable_tuple] = {}

            # Set the x variable (i.e. "x" or "r" for any one-dimensional variables)
            if first_variable.dimensions == 1:
                (spatial_var_name, spatial_var_value) = self._get_spatial_var(
                    variable_tuple, first_variable, "first"
                )
                self.spatial_variable_dict[variable_tuple] = {
                    spatial_var_name: spatial_var_value
                }
                self.first_spatial_variable[variable_tuple] = (
                    spatial_var_value * self.spatial_factor
                )

            elif first_variable.dimensions == 2:
                # Don't allow 2D variables if there are multiple solutions
                if len(variables) > 1:
                    raise NotImplementedError(
                        "Cannot plot 2D variables when comparing multiple solutions, "
                        f"but '{variable_tuple[0]}' is 2D"
                    )
                # But do allow if just a single solution
                else:
                    # Add both spatial variables to the variable_tuples
                    (
                        first_spatial_var_name,
                        first_spatial_var_value,
                    ) = self._get_spatial_var(variable_tuple, first_variable, "first")
                    (
                        second_spatial_var_name,
                        second_spatial_var_value,
                    ) = self._get_spatial_var(variable_tuple, first_variable, "second")
                    self.spatial_variable_dict[variable_tuple] = {
                        first_spatial_var_name: first_spatial_var_value,
                        second_spatial_var_name: second_spatial_var_value,
                    }
                    self.first_spatial_variable[variable_tuple] = (
                        first_spatial_var_value * self.spatial_factor
                    )
                    self.second_spatial_variable[variable_tuple] = (
                        second_spatial_var_value * self.spatial_factor
                    )
                    # different order based on whether the domains
                    # are x-r, x-z or y-z, etc
                    if (
                        first_spatial_var_name in ("r", "R")
                        and second_spatial_var_name == "x"
                    ):
                        self.x_first_and_y_second[variable_tuple] = False
                        self.is_y_z[variable_tuple] = False
                    elif (
                        first_spatial_var_name == "y" and second_spatial_var_name == "z"
                    ):
                        self.x_first_and_y_second[variable_tuple] = True
                        self.is_y_z[variable_tuple] = True
                    else:
                        self.x_first_and_y_second[variable_tuple] = True
                        self.is_y_z[variable_tuple] = False

            # Store variables and subplot position
            self.variables[variable_tuple] = variables
            self.subplot_positions[variable_tuple] = (self.n_rows, self.n_cols, k + 1)

    def _get_spatial_var(self, key, variable, dimension):
        """Return the appropriate spatial variable(s)"""

        # Extract name and value
        # Special case for current collector, which is 2D but in a weird way (both
        # first and second variables are in the same domain, not auxiliary domain)
        if dimension == "first":
            spatial_var_name = variable.first_dimension
            spatial_var_value = variable.first_dim_pts
            domain = variable.domain[0]
        elif dimension == "second":
            spatial_var_name = variable.second_dimension
            spatial_var_value = variable.second_dim_pts
            if variable.domain[0] == "current collector":
                domain = "current collector"
            elif isinstance(variable, pybamm.ProcessedVariable2DFVM):
                domain = variable.domain[0]
            else:
                domain = variable.domains["secondary"][0]

        if domain == "current collector":
            domain += f" {spatial_var_name}"

        return spatial_var_name, spatial_var_value

    def reset_axis(self):
        """
        Reset the axis limits to the default values.
        These are calculated to fit around the minimum and maximum values of all the
        variables in each subplot
        """
        self.axis_limits = {}
        for key, variable_lists in self.variables.items():
            if variable_lists[0][0].dimensions == 0:
                x_min = self.min_t
                x_max = self.max_t
            elif variable_lists[0][0].dimensions == 1:
                x_min = self.first_spatial_variable[key][0]
                x_max = self.first_spatial_variable[key][-1]
            elif variable_lists[0][0].dimensions == 2:
                # different order based on whether the domains are x-r, x-z or y-z, etc
                if self.x_first_and_y_second[key] is False:
                    x_min = self.second_spatial_variable[key][0]
                    x_max = self.second_spatial_variable[key][-1]
                    y_min = self.first_spatial_variable[key][0]
                    y_max = self.first_spatial_variable[key][-1]
                else:
                    x_min = self.first_spatial_variable[key][0]
                    x_max = self.first_spatial_variable[key][-1]
                    y_min = self.second_spatial_variable[key][0]
                    y_max = self.second_spatial_variable[key][-1]

                # Create axis for contour plot
                self.axis_limits[key] = [x_min, x_max, y_min, y_max]

            # Get min and max variable values
            if self.variable_limits[key] == "fixed":
                # fixed variable limits: calculate "globlal" min and max
                spatial_vars = self.spatial_variable_dict[key]
                var_min = np.min(
                    [
                        ax_min(var(self.ts_seconds[i], **spatial_vars))
                        for i, variable_list in enumerate(variable_lists)
                        for var in variable_list
                    ]
                )
                var_max = np.max(
                    [
                        ax_max(var(self.ts_seconds[i], **spatial_vars))
                        for i, variable_list in enumerate(variable_lists)
                        for var in variable_list
                    ]
                )
                if np.isnan(var_min) or np.isnan(var_max):
                    raise ValueError(
                        "The variable limits are set to 'fixed' but the min and max "
                        "values are NaN"
                    )
                if var_min == var_max:
                    var_min -= 1
                    var_max += 1
            elif self.variable_limits[key] == "tight":
                # tight variable limits: axes will adjust each time
                var_min, var_max = None, None
            else:
                # user-specified axis limits
                var_min, var_max = self.variable_limits[key]

            if variable_lists[0][0].dimensions in [0, 1]:
                self.axis_limits[key] = [x_min, x_max, var_min, var_max]
            else:
                self.variable_limits[key] = (var_min, var_max)
            if (
                var_min is not None
                and var_max is not None
                and (np.isnan(var_min) or np.isnan(var_max))
            ):  # pragma: no cover
                raise ValueError(f"Axis limits cannot be NaN for variables '{key}'")

    def plot(self, t, dynamic=False):
        """Produces a quick plot with the internal states at time t.

        Parameters
        ----------
        t : float
            Dimensional time (in 'time_units') at which to plot.
        dynamic : bool, optional
            Determine whether to allocate space for a slider at the bottom of the plot when generating a dynamic plot.
            If True, creates a dynamic plot with a slider.
        """

        plt = import_optional_dependency("matplotlib.pyplot")
        gridspec = import_optional_dependency("matplotlib.gridspec")
        cm = import_optional_dependency("matplotlib", "cm")
        colors = import_optional_dependency("matplotlib", "colors")

        t_in_seconds = t * self.time_scaling_factor
        t_in_seconds = np.clip(t_in_seconds, self.min_t_unscaled, self.max_t_unscaled)
        self.fig = plt.figure(figsize=self.figsize)

        self.gridspec = gridspec.GridSpec(self.n_rows, self.n_cols)
        self.plots = {}
        self.time_lines = {}
        self.colorbars = {}
        self.axes = QuickPlotAxes()

        # initialize empty handles, to be created only if the appropriate plots are made
        solution_handles = []

        for k, (key, variable_lists) in enumerate(self.variables.items()):
            ax = self.fig.add_subplot(self.gridspec[k])
            self.axes.add(key, ax)
            x_min, x_max, y_min, y_max = self.axis_limits[key]
            ax.set_xlim(x_min, x_max)
            if y_min is not None and y_max is not None:
                ax.set_ylim(y_min, y_max)
            ax.xaxis.set_major_locator(plt.MaxNLocator(3))
            self.plots[key] = defaultdict(dict)
            variable_handles = []
            # Set labels for the first subplot only (avoid repetition)
            if variable_lists[0][0].dimensions == 0:
                # 0D plot: plot as a function of time, indicating time t with a line
                ax.set_xlabel(f"Time [{self.time_unit}]")
                for i, variable_list in enumerate(variable_lists):
                    for j, variable in enumerate(variable_list):
                        if len(variable_list) == 1:
                            # single variable -> use linestyle to differentiate model
                            linestyle = self.linestyles[i]
                        else:
                            # multiple variables -> use linestyle to differentiate
                            # variables (color differentiates models)
                            linestyle = self.linestyles[j]
                        full_t = self.ts_seconds[i]
                        (self.plots[key][i][j],) = ax.plot(
                            full_t / self.time_scaling_factor,
                            variable(full_t),
                            color=self.colors[i],
                            linestyle=linestyle,
                        )
                        variable_handles.append(self.plots[key][0][j])
                    solution_handles.append(self.plots[key][i][0])
                y_min, y_max = ax.get_ylim()
                ax.set_ylim(y_min, y_max)
                (self.time_lines[key],) = ax.plot(
                    [t, t],
                    [y_min, y_max],
                    "k--",
                    lw=1.5,
                )
            elif variable_lists[0][0].dimensions == 1:
                # 1D plot: plot as a function of x at time t
                # Read dictionary of spatial variables
                spatial_vars = self.spatial_variable_dict[key]
                spatial_var_name = next(iter(spatial_vars.keys()))
                ax.set_xlabel(
                    f"{spatial_var_name} [{self.spatial_unit}]",
                )
                for i, variable_list in enumerate(variable_lists):
                    for j, variable in enumerate(variable_list):
                        if len(variable_list) == 1:
                            # single variable -> use linestyle to differentiate model
                            linestyle = self.linestyles[i]
                        else:
                            # multiple variables -> use linestyle to differentiate
                            # variables (color differentiates models)
                            linestyle = self.linestyles[j]
                        (self.plots[key][i][j],) = ax.plot(
                            self.first_spatial_variable[key],
                            variable(t_in_seconds, **spatial_vars),
                            color=self.colors[i],
                            linestyle=linestyle,
                            zorder=10,
                        )
                        variable_handles.append(self.plots[key][0][j])
                    solution_handles.append(self.plots[key][i][0])
                # add lines for boundaries between subdomains
                for boundary in variable_lists[0][0].internal_boundaries:
                    boundary_scaled = boundary * self.spatial_factor
                    ax.axvline(boundary_scaled, color="0.5", lw=1, zorder=0)
            elif variable_lists[0][0].dimensions == 2:
                # Read dictionary of spatial variables
                spatial_vars = self.spatial_variable_dict[key]
                # there can only be one entry in the variable list
                variable = variable_lists[0][0]
                # different order based on whether the domains are x-r, x-z or y-z, etc
                if self.x_first_and_y_second[key] is False:
                    x_name = list(spatial_vars.keys())[1][0]
                    y_name = next(iter(spatial_vars.keys()))[0]
                    x = self.second_spatial_variable[key]
                    y = self.first_spatial_variable[key]
                    var = variable(t_in_seconds, **spatial_vars)
                else:
                    x_name = next(iter(spatial_vars.keys()))[0]
                    y_name = list(spatial_vars.keys())[1][0]
                    x = self.first_spatial_variable[key]
                    y = self.second_spatial_variable[key]
                    var = variable(t_in_seconds, **spatial_vars).T
                ax.set_xlabel(f"{x_name} [{self.spatial_unit}]")
                ax.set_ylabel(f"{y_name} [{self.spatial_unit}]")
                vmin, vmax = self.variable_limits[key]
                # store the plot and the var data (for testing) as cant access
                # z data from QuadMesh or QuadContourSet object
                if self.is_y_z[key] is True:
                    self.plots[key][0][0] = ax.pcolormesh(
                        x,
                        y,
                        var,
                        vmin=vmin,
                        vmax=vmax,
                        shading=self.shading,
                    )
                else:
                    self.plots[key][0][0] = ax.contourf(
                        x, y, var, levels=100, vmin=vmin, vmax=vmax
                    )
                self.plots[key][0][1] = var
                if vmin is None and vmax is None:
                    vmin = ax_min(var)
                    vmax = ax_max(var)
                self.colorbars[key] = self.fig.colorbar(
                    cm.ScalarMappable(colors.Normalize(vmin=vmin, vmax=vmax)),
                    ax=ax,
                )
            # Set either y label or legend entries
            if len(key) == 1:
                title = split_long_string(key[0])
                ax.set_title(title, fontsize="medium")
            else:
                ax.legend(
                    variable_handles,
                    [split_long_string(s, 6) for s in key],
                    bbox_to_anchor=(0.5, 1),
                    loc="lower center",
                )

        # Set global legend
        if len(self.labels) > 1:
            fig_legend = self.fig.legend(
                solution_handles, self.labels, loc="lower right"
            )
            # Get the position of the top of the legend in relative figure units
            # There may be a better way ...
            try:
                legend_top_inches = fig_legend.get_window_extent(
                    renderer=self.fig.canvas.get_renderer()
                ).get_points()[1, 1]
                fig_height_inches = (self.fig.get_size_inches() * self.fig.dpi)[1]
                legend_top = legend_top_inches / fig_height_inches
            except AttributeError:  # pragma: no cover
                # When testing the examples we set the matplotlib backend to "Template"
                # which means that the above code doesn't work. Since this is just for
                # that particular test we can just skip it
                legend_top = 0
        else:
            legend_top = 0

        # Fix layout
        if dynamic:
            slider_top = 0.05
        else:
            slider_top = 0
        bottom = max(legend_top, slider_top)
        self.gridspec.tight_layout(self.fig, rect=[0, bottom, 1, 1])

    def dynamic_plot(self, show_plot=True, step=None):
        """
        Generate a dynamic plot with a slider to control the time.

        Parameters
        ----------
        step : float, optional
            For notebook mode, size of steps to allow in the slider. Defaults to 1/100th
            of the total time.
        show_plot : bool, optional
            Whether to show the plots. Default is True. Set to False if you want to
            only display the plot after plt.show() has been called.

        """
        if pybamm.is_notebook():  # pragma: no cover
            import ipywidgets as widgets

            step = step or self.max_t / 100
            widgets.interact(
                lambda t: self.plot(t, dynamic=False),
                t=widgets.FloatSlider(
                    min=self.min_t, max=self.max_t, step=step, value=self.min_t
                ),
                continuous_update=False,
            )
        else:
            plt = import_optional_dependency("matplotlib.pyplot")
            Slider = import_optional_dependency("matplotlib.widgets", "Slider")

            # create an initial plot at time self.min_t
            self.plot(self.min_t, dynamic=True)

            axcolor = "lightgoldenrodyellow"
            ax_slider = plt.axes([0.315, 0.02, 0.37, 0.03], facecolor=axcolor)
            self.slider = Slider(
                ax_slider,
                f"Time [{self.time_unit}]",
                self.min_t,
                self.max_t,
                valinit=self.min_t,
                color="#1f77b4",
            )
            self.slider.on_changed(self.slider_update)

            if show_plot:  # pragma: no cover
                plt.show()

    def slider_update(self, t):
        """
        Update the plot in self.plot() with values at new time
        """
        from matplotlib import cm, colors

        time_in_seconds = t * self.time_scaling_factor
        time_in_seconds = np.clip(
            time_in_seconds, self.min_t_unscaled, self.max_t_unscaled
        )
        for k, (key, plot) in enumerate(self.plots.items()):
            ax = self.axes[k]
            if self.variables[key][0][0].dimensions == 0:
                self.time_lines[key].set_xdata([t])
            elif self.variables[key][0][0].dimensions == 1:
                var_min = np.inf
                var_max = -np.inf
                for i, variable_lists in enumerate(self.variables[key]):
                    for j, variable in enumerate(variable_lists):
                        var = variable(
                            time_in_seconds,
                            **self.spatial_variable_dict[key],
                        )
                        plot[i][j].set_ydata(var)
                        var_min = min(var_min, ax_min(var))
                        var_max = max(var_max, ax_max(var))
                # update boundaries between subdomains
                y_min, y_max = self.axis_limits[key][2:]
                if y_min is None and y_max is None:
                    ax.set_ylim(var_min, var_max)
            elif self.variables[key][0][0].dimensions == 2:
                # 2D plot: plot as a function of x and y at time t
                # Read dictionary of spatial variables
                spatial_vars = self.spatial_variable_dict[key]
                # there can only be one entry in the variable list
                variable = self.variables[key][0][0]
                vmin, vmax = self.variable_limits[key]
                if self.x_first_and_y_second[key] is False:
                    x = self.second_spatial_variable[key]
                    y = self.first_spatial_variable[key]
                    var = variable(time_in_seconds, **spatial_vars)
                else:
                    x = self.first_spatial_variable[key]
                    y = self.second_spatial_variable[key]
                    var = variable(time_in_seconds, **spatial_vars).T
                # store the plot and the var data (for testing) as cant access
                # z data from QuadMesh or QuadContourSet object
                if self.is_y_z[key] is True:
                    self.plots[key][0][0] = ax.pcolormesh(
                        x,
                        y,
                        var,
                        vmin=vmin,
                        vmax=vmax,
                        shading=self.shading,
                    )
                else:
                    self.plots[key][0][0] = ax.contourf(
                        x, y, var, levels=100, vmin=vmin, vmax=vmax
                    )
                self.plots[key][0][1] = var
                if (vmin, vmax) == (None, None):
                    vmin = ax_min(var)
                    vmax = ax_max(var)
                    cb = self.colorbars[key]
                    cb.update_normal(
                        cm.ScalarMappable(colors.Normalize(vmin=vmin, vmax=vmax))
                    )

        self.fig.canvas.draw_idle()

    def create_gif(self, number_of_images=80, duration=0.1, output_filename="plot.gif"):
        """
        Generates x plots over a time span of max_t - min_t and compiles them to create
        a GIF.

        Parameters
        ----------
        number_of_images : int, optional
            Number of images/plots to be compiled for a GIF.
        duration : float, optional
            Duration of visibility of a single image/plot in the created GIF.
        output_filename : str, optional
            Name of the generated GIF file.

        """
        FuncAnimation = import_optional_dependency(
            "matplotlib.animation", "FuncAnimation"
        )

        # time stamps at which the images/plots will be created
        time_array = np.linspace(self.min_t, self.max_t, num=number_of_images)

        # create images/plots
        self.plot(self.min_t)
        ani = FuncAnimation(
            self.fig,
            lambda f: self.slider_update(time_array[f - 1]),
            interval=duration,
            frames=number_of_images,
        )
        ani.save(output_filename, dpi=300)

    @property
    def min_t(self):
        return self.min_t_unscaled / self.time_scaling_factor

    @property
    def max_t(self):
        return self.max_t_unscaled / self.time_scaling_factor


class QuickPlotAxes:
    """
    Class to store axes for the QuickPlot
    """

    def __init__(self):
        self._by_variable = {}
        self._axes = []

    def add(self, keys, axis):
        """
        Add axis

        Parameters
        ----------
        keys : iter
            Iterable of keys of variables being plotted on the axis
        axis : matplotlib Axis object
            The axis object
        """
        self._axes.append(axis)
        for k in keys:
            self._by_variable[k] = axis

    def __getitem__(self, index):
        """
        Get axis by index
        """
        return self._axes[index]

    def by_variable(self, key):
        """
        Get axis by variable name
        """
        return self._by_variable[key]



#
# Lithium-ion base model class
#
import pybamm


class BaseModel(pybamm.BaseBatteryModel):
    """
    Overwrites default parameters from Base Model with default parameters for
    lithium-ion models

    Parameters
    ----------
    options : dict-like, optional
        A dictionary of options to be passed to the model. If this is a dict (and not
        a subtype of dict), it will be processed by :class:`pybamm.BatteryModelOptions`
        to ensure that the options are valid. If this is a subtype of dict, it is
        assumed that the options have already been processed and are valid. This allows
        for the use of custom options classes. The default options are given by
        :class:`pybamm.BatteryModelOptions`.
    name : str, optional
        The name of the model. The default is "Unnamed battery model".
    build : bool, optional
        Whether to build the model on instantiation. Default is True. Setting this
        option to False allows users to change any number of the submodels before
        building the complete model (submodels cannot be changed after the model is
        built).
    """

    def __init__(self, options=None, name="Unnamed lithium-ion model", build=False):
        super().__init__(options, name)
        self.param = pybamm.LithiumIonParameters(self.options)

        self.set_standard_output_variables()

        # Li models should calculate esoh by default
        self._calc_esoh = True

    def set_submodels(self, build):
        self.set_external_circuit_submodel()
        self.set_porosity_submodel()
        self.set_interface_utilisation_submodel()
        self.set_crack_submodel()
        self.set_active_material_submodel()
        self.set_transport_efficiency_submodels()
        self.set_convection_submodel()
        self.set_open_circuit_potential_submodel()
        self.set_intercalation_kinetics_submodel()
        self.set_particle_submodel()
        self.set_solid_submodel()
        self.set_electrolyte_concentration_submodel()
        self.set_electrolyte_potential_submodel()
        self.set_thermal_submodel()
        self.set_surface_temperature_submodel()
        self.set_current_collector_submodel()
        self.set_sei_submodel()
        self.set_sei_on_cracks_submodel()
        self.set_lithium_plating_submodel()
        self.set_li_metal_counter_electrode_submodels()
        self.set_total_interface_submodel()

        if build:
            self.build_model()

    @property
    def default_parameter_values(self):
        if self.options.whole_cell_domains == [
            "negative electrode",
            "separator",
            "positive electrode",
        ]:
            return pybamm.ParameterValues("Marquis2019")
        else:
            return pybamm.ParameterValues("Xu2019")

    @property
    def default_quick_plot_variables(self):
        if self.options.whole_cell_domains == ["separator", "positive electrode"]:
            return [
                "Electrolyte concentration [mol.m-3]",
                "Positive particle surface concentration [mol.m-3]",
                "Current [A]",
                "Electrolyte potential [V]",
                "Positive electrode potential [V]",
                "Voltage [V]",
            ]
        else:
            return [
                "Negative particle surface concentration [mol.m-3]",
                "Electrolyte concentration [mol.m-3]",
                "Positive particle surface concentration [mol.m-3]",
                "Current [A]",
                "Negative electrode potential [V]",
                "Electrolyte potential [V]",
                "Positive electrode potential [V]",
                "Voltage [V]",
            ]

    @property
    def calc_esoh(self):
        """Whether to include eSOH variables in the summary variables."""
        if (
            self.options["particle phases"] not in ["1", ("1", "1")]
            or self.options["working electrode"] != "both"
        ):
            self._calc_esoh = False
        return self._calc_esoh

    @calc_esoh.setter
    def calc_esoh(self, value: bool):
        """Set whether to include eSOH variables in the summary variables."""
        if not isinstance(value, bool):
            raise TypeError("`calc_esoh` arg needs to be a bool")

        self._calc_esoh = value

    def set_standard_output_variables(self):
        super().set_standard_output_variables()

        # Particle concentration position
        var = pybamm.standard_spatial_vars
        if self.options.electrode_types["negative"] == "porous":
            self.variables.update({"r_n [m]": var.r_n})
        if self.options.electrode_types["positive"] == "porous":
            self.variables.update({"r_p [m]": var.r_p})

    def set_degradation_variables(self):
        """Sets variables that quantify degradation (LAM, LLI, etc)"""

        domains = [d for d in self.options.whole_cell_domains if d != "separator"]
        for domain in domains:
            Domain = domain.capitalize()
            self.variables[f"Total lithium in {domain} [mol]"] = sum(
                self.variables[f"Total lithium in {phase} phase in {domain} [mol]"]
                for phase in self.options.phases[domain.split()[0]]
            )

            # LAM
            Q_k = self.variables[f"{Domain} capacity [A.h]"]
            domain_param = getattr(self.param, domain[0])  # param.n or param.p
            LAM_k = (1 - Q_k / domain_param.Q_init) * 100
            self.variables.update(
                {
                    f"LAM_{domain[0]}e [%]": LAM_k,
                    f"Loss of active material in {domain} [%]": LAM_k,
                }
            )

        # LLI
        n_Li_e = self.variables["Total lithium in electrolyte [mol]"]
        n_Li_particles = sum(
            self.variables[f"Total lithium in {domain} [mol]"] for domain in domains
        )
        n_Li = n_Li_particles + n_Li_e

        # LLI is usually defined based only on the percentage lithium lost from
        # particles
        LLI = (1 - n_Li_particles / self.param.n_Li_particles_init) * 100
        LLI_tot = (1 - n_Li / self.param.n_Li_init) * 100

        self.variables.update(
            {
                "LLI [%]": LLI,
                "Loss of lithium inventory [%]": LLI,
                "Loss of lithium inventory, including electrolyte [%]": LLI_tot,
                # Total lithium
                "Total lithium [mol]": n_Li,
                "Total lithium in particles [mol]": n_Li_particles,
                "Total lithium capacity [A.h]": n_Li * self.param.F / 3600,
                "Total lithium capacity in particles [A.h]": n_Li_particles
                * self.param.F
                / 3600,
                # Lithium lost
                "Total lithium lost [mol]": self.param.n_Li_init - n_Li,
                "Total lithium lost from particles [mol]": self.param.n_Li_particles_init
                - n_Li_particles,
                "Total lithium lost from electrolyte [mol]": self.param.n_Li_e_init
                - n_Li_e,
            }
        )

        # Lithium lost to side reactions
        # Different way of measuring LLI but should give same value
        n_Li_lost_neg_sei = self.variables["Loss of lithium to negative SEI [mol]"]
        n_Li_lost_pos_sei = self.variables["Loss of lithium to positive SEI [mol]"]
        n_Li_lost_reactions = n_Li_lost_neg_sei + n_Li_lost_pos_sei
        for domain in domains:
            dom = domain.split()[0].lower()
            n_Li_lost_sei_cracks = self.variables[
                f"Loss of lithium to {dom} SEI on cracks [mol]"
            ]
            n_Li_lost_pl = self.variables[
                f"Loss of lithium to {dom} lithium plating [mol]"
            ]
            n_Li_lost_reactions += n_Li_lost_sei_cracks + n_Li_lost_pl

        self.variables.update(
            {
                "Total lithium lost to side reactions [mol]": n_Li_lost_reactions,
                "Total capacity lost to side reactions [A.h]": n_Li_lost_reactions
                * self.param.F
                / 3600,
            }
        )

    def set_default_summary_variables(self):
        """
        Sets the default summary variables.
        """
        summary_variables = [
            "Time [s]",
            "Time [h]",
            "Throughput capacity [A.h]",
            "Throughput energy [W.h]",
            # LAM, LLI
            "Loss of lithium inventory [%]",
            "Loss of lithium inventory, including electrolyte [%]",
            # Total lithium
            "Total lithium [mol]",
            "Total lithium in electrolyte [mol]",
            "Total lithium in particles [mol]",
            # Lithium lost
            "Total lithium lost [mol]",
            "Total lithium lost from particles [mol]",
            "Total lithium lost from electrolyte [mol]",
            "Loss of lithium to negative SEI [mol]",
            "Loss of capacity to negative SEI [A.h]",
            "Loss of lithium to positive SEI [mol]",
            "Loss of capacity to positive SEI [A.h]",
            "Total lithium lost to side reactions [mol]",
            "Total capacity lost to side reactions [A.h]",
            # Resistance
            "Local ECM resistance [Ohm]",
        ]

        if self.options.electrode_types["negative"] == "porous":
            summary_variables += [
                "Negative electrode capacity [A.h]",
                "Loss of active material in negative electrode [%]",
                "Total lithium in negative electrode [mol]",
                "Loss of lithium to negative lithium plating [mol]",
                "Loss of capacity to negative lithium plating [A.h]",
                "Loss of lithium to negative SEI on cracks [mol]",
                "Loss of capacity to negative SEI on cracks [A.h]",
            ]
        if self.options.electrode_types["positive"] == "porous":
            summary_variables += [
                "Positive electrode capacity [A.h]",
                "Loss of active material in positive electrode [%]",
                "Total lithium in positive electrode [mol]",
                "Loss of lithium to positive lithium plating [mol]",
                "Loss of capacity to positive lithium plating [A.h]",
                "Loss of lithium to positive SEI on cracks [mol]",
                "Loss of capacity to positive SEI on cracks [A.h]",
            ]

        self.summary_variables = summary_variables

    def set_open_circuit_potential_submodel(self):
        for domain in ["negative", "positive"]:
            if self.options.electrode_types[domain] == "porous":
                reaction = "lithium-ion main"
            elif self.options.electrode_types[domain] == "planar":
                reaction = "lithium metal plating"
            domain_options = getattr(self.options, domain)
            for phase in self.options.phases[domain]:
                ocp_option = getattr(domain_options, phase)["open-circuit potential"]
                ocp_submodels = pybamm.open_circuit_potential
                if ocp_option == "single":
                    ocp_model = ocp_submodels.SingleOpenCircuitPotential
                elif ocp_option == "current sigmoid":
                    ocp_model = ocp_submodels.CurrentSigmoidOpenCircuitPotential
                elif ocp_option == "one-state hysteresis":
                    ocp_model = ocp_submodels.OneStateHysteresisOpenCircuitPotential
                elif ocp_option == "one-state differential capacity hysteresis":
                    ocp_model = ocp_submodels.OneStateDifferentialCapacityHysteresisOpenCircuitPotential
                elif ocp_option == "MSMR":
                    ocp_model = ocp_submodels.MSMROpenCircuitPotential
                self.submodels[f"{domain} {phase} open-circuit potential"] = ocp_model(
                    self.param, domain, reaction, self.options, phase, self.x_average
                )

    def set_sei_submodel(self):
        for domain in ["negative", "positive"]:
            phases = self.options.phases[domain]
            for phase in phases:
                sei_option = getattr(getattr(self.options, domain), phase)["SEI"]
                if sei_option == "none":
                    submodel = pybamm.sei.NoSEI(self.param, domain, self.options, phase)
                elif sei_option == "constant":
                    submodel = pybamm.sei.ConstantSEI(
                        self.param, domain, self.options, phase
                    )
                else:
                    submodel = pybamm.sei.SEIGrowth(
                        self.param,
                        domain,
                        self.options,
                        phase,
                        cracks=False,
                    )
                self.submodels[f"{domain} {phase} sei"] = submodel
                self.submodels[f"{domain} {phase} sei thickness"] = (
                    pybamm.sei.SEIThickness(
                        self.param, domain, self.options, phase, cracks=False
                    )
                )
            if len(phases) > 1:
                self.submodels[f"{domain} total sei"] = pybamm.sei.TotalSEI(
                    self.param, domain, self.options
                )

    def set_sei_on_cracks_submodel(self):
        # Do not set "sei on cracks" submodel for a planar electrode. For porous
        # electrodes, "sei on cracks" submodel must be set, even if it is zero
        for domain in self.options.whole_cell_domains:
            if domain != "separator":
                domain = domain.split()[0].lower()
                sei_option = getattr(self.options, domain)["SEI"]
                sei_on_cracks_option = getattr(self.options, domain)["SEI on cracks"]
                phases = self.options.phases[domain]
                for phase in phases:
                    if (
                        sei_option in ["none", "constant"]
                        or sei_on_cracks_option == "false"
                    ):
                        submodel = pybamm.sei.NoSEI(
                            self.param, domain, self.options, phase, cracks=True
                        )
                    else:
                        submodel = pybamm.sei.SEIGrowth(
                            self.param,
                            domain,
                            self.options,
                            phase,
                            cracks=True,
                        )
                    self.submodels[f"{domain} {phase} sei on cracks"] = submodel
                    self.submodels[f"{domain} {phase} sei on cracks thickness"] = (
                        pybamm.sei.SEIThickness(
                            self.param,
                            domain,
                            self.options,
                            phase,
                            cracks=True,
                        )
                    )
                if len(phases) > 1:
                    self.submodels[f"{domain} total sei on cracks"] = (
                        pybamm.sei.TotalSEI(
                            self.param, domain, self.options, cracks=True
                        )
                    )

    def set_lithium_plating_submodel(self):
        # Do not set "lithium plating" submodel for a planar electrode. For porous
        # electrodes, "lithium plating" submodel must be set, even if it is zero
        for domain in self.options.whole_cell_domains:
            if domain != "separator":
                domain = domain.split()[0].lower()
                phases = self.options.phases[domain]
                for phase in phases:
                    lithium_plating_opt = getattr(getattr(self.options, domain), phase)[
                        "lithium plating"
                    ]
                    if lithium_plating_opt == "none":
                        submodel = pybamm.lithium_plating.NoPlating(
                            self.param, domain, self.options, phase
                        )
                    else:
                        x_average = self.options["x-average side reactions"] == "true"
                        submodel = pybamm.lithium_plating.Plating(
                            self.param, domain, x_average, self.options, phase
                        )
                    self.submodels[f"{domain} {phase} lithium plating"] = submodel
                if len(phases) > 1:
                    self.submodels[f"{domain} total lithium plating"] = (
                        pybamm.lithium_plating.TotalLithiumPlating(
                            self.param, domain, self.options
                        )
                    )

    def set_total_interface_submodel(self):
        self.submodels["total interface"] = pybamm.interface.TotalInterfacialCurrent(
            self.param, "lithium-ion", self.options
        )

    def set_crack_submodel(self):
        for domain in self.options.whole_cell_domains:
            if domain != "separator":
                domain = domain.split()[0].lower()
                phases = self.options.phases[domain]
                for phase in phases:
                    crack = getattr(self.options, domain)["particle mechanics"]
                    if crack == "none":
                        self.submodels[f"{domain} {phase}particle mechanics"] = (
                            pybamm.particle_mechanics.NoMechanics(
                                self.param, domain, options=self.options, phase=phase
                            )
                        )
                    elif crack == "swelling only":
                        self.submodels[f"{domain} {phase}particle mechanics"] = (
                            pybamm.particle_mechanics.SwellingOnly(
                                self.param, domain, options=self.options, phase=phase
                            )
                        )
                    elif crack == "swelling and cracking":
                        self.submodels[f"{domain} {phase}particle mechanics"] = (
                            pybamm.particle_mechanics.CrackPropagation(
                                self.param,
                                domain,
                                self.x_average,
                                options=self.options,
                                phase=phase,
                            )
                        )

    def set_active_material_submodel(self):
        for domain in ["negative", "positive"]:
            if self.options.electrode_types[domain] == "porous":
                lam = getattr(self.options, domain)["loss of active material"]
                phases = self.options.phases[domain]
                for phase in phases:
                    if lam == "none":
                        submod = pybamm.active_material.Constant(
                            self.param, domain, self.options, phase
                        )
                    else:
                        submod = pybamm.active_material.LossActiveMaterial(
                            self.param, domain, self.options, self.x_average, phase
                        )
                    self.submodels[f"{domain} {phase} active material"] = submod

                # Submodel for the total active material, summing up each phase
                if len(phases) > 1:
                    self.submodels[f"{domain} total active material"] = (
                        pybamm.active_material.Total(self.param, domain, self.options)
                    )

    def set_porosity_submodel(self):
        if (
            self.options["SEI porosity change"] == "false"
            and self.options["lithium plating porosity change"] == "false"
        ):
            self.submodels["porosity"] = pybamm.porosity.Constant(
                self.param, self.options
            )
        elif (
            self.options["SEI porosity change"] == "true"
            or self.options["lithium plating porosity change"] == "true"
        ):
            x_average = self.options["x-average side reactions"] == "true"
            self.submodels["porosity"] = pybamm.porosity.ReactionDriven(
                self.param, self.options, x_average
            )

    def set_li_metal_counter_electrode_submodels(self):
        for domain in ["negative", "positive"]:
            if self.options.electrode_types[domain] == "porous":
                continue
            if (
                self.options["SEI"] in ["none", "constant"]
                and self.options["intercalation kinetics"] == "symmetric Butler-Volmer"
                and self.options["surface form"] == "false"
            ):
                # only symmetric Butler-Volmer can be inverted
                self.submodels[f"{domain} electrode potential"] = (
                    pybamm.electrode.ohm.LithiumMetalExplicit(
                        self.param, domain, self.options
                    )
                )
                self.submodels[f"{domain} electrode interface"] = (
                    pybamm.kinetics.InverseButlerVolmer(
                        self.param, domain, "lithium metal plating", self.options
                    )
                )  # assuming symmetric reaction for now so we can take the inverse
                self.submodels[f"{domain} electrode interface current"] = (
                    pybamm.kinetics.CurrentForInverseButlerVolmerLithiumMetal(
                        self.param, domain, "lithium metal plating", self.options
                    )
                )
            else:
                self.submodels[f"{domain} electrode potential"] = (
                    pybamm.electrode.ohm.LithiumMetalSurfaceForm(
                        self.param, domain, self.options
                    )
                )
                neg_intercalation_kinetics = self.get_intercalation_kinetics(domain)
                self.submodels[f"{domain} electrode interface"] = (
                    neg_intercalation_kinetics(
                        self.param,
                        domain,
                        "lithium metal plating",
                        self.options,
                        "primary",
                    )
                )

    def set_convection_submodel(self):
        self.submodels["transverse convection"] = (
            pybamm.convection.transverse.NoConvection(self.param, self.options)
        )
        self.submodels["through-cell convection"] = (
            pybamm.convection.through_cell.NoConvection(self.param, self.options)
        )

    def insert_reference_electrode(self, position=None):
        """
        Insert a reference electrode to measure the electrolyte potential at a given
        position in space. Adds model variables for the electrolyte potential at the
        reference electrode and for the potential difference between the electrode
        potentials measured at the electrode/current collector interface and the
        reference electrode. Only implemented for 1D models (i.e. where the
        'dimensionality' option is 0).

        Parameters
        ----------
        position : :class:`pybamm.Symbol`, optional
            The position in space at which to measure the electrolyte potential. If
            None, defaults to the mid-point of the separator.
        """
        if self.options["dimensionality"] != 0:
            raise NotImplementedError(
                "Reference electrode can only be inserted for models where "
                "'dimensionality' is 0. For other models, please add a reference "
                "electrode manually."
            )

        if position is None:
            position = self.param.n.L + self.param.s.L / 2

        phi_e_ref = pybamm.EvaluateAt(
            self.variables["Electrolyte potential [V]"], position
        )
        phi_p = pybamm.boundary_value(
            self.variables["Positive electrode potential [V]"], "right"
        )
        variables = {
            "Positive electrode 3E potential [V]": phi_p - phi_e_ref,
            "Reference electrode potential [V]": phi_e_ref,
        }

        if self.options["working electrode"] == "both":
            phi_n = pybamm.boundary_value(
                self.variables["Negative electrode potential [V]"], "left"
            )
            variables.update(
                {
                    "Negative electrode 3E potential [V]": phi_n - phi_e_ref,
                }
            )
        self.variables.update(variables)



#
# Standard electrical parameters
#
import pybamm

from .base_parameters import BaseParameters


class ElectricalParameters(BaseParameters):
    """
    Standard electrical parameters
    """

    def __init__(self):
        # Get geometric parameters
        self.geo = pybamm.geometric_parameters

        # Set parameters
        self._set_parameters()

    def _set_parameters(self):
        """Defines the dimensional parameters."""

        self.Q = pybamm.Parameter("Nominal cell capacity [A.h]")
        self.R_contact = pybamm.Parameter("Contact resistance [Ohm]")
        self.n_cells = pybamm.Parameter(
            "Number of cells connected in series to make a battery"
        )
        self.voltage_low_cut = pybamm.Parameter("Lower voltage cut-off [V]")
        self.voltage_high_cut = pybamm.Parameter("Upper voltage cut-off [V]")
        self.ocp_soc_0 = pybamm.Parameter("Open-circuit voltage at 0% SOC [V]")
        self.ocp_soc_100 = pybamm.Parameter("Open-circuit voltage at 100% SOC [V]")
        # Current as a function of time
        self.current_with_time = pybamm.FunctionParameter(
            "Current function [A]", {"Time [s]": pybamm.t}
        )
        self.current_density_with_time = self.current_with_time / (self.geo.A_cc)


electrical_parameters = ElectricalParameters()



#
# Class for leading-order surface form electrolyte conductivity employing stefan-maxwell
#
import pybamm
from pybamm.models.submodels.electrolyte_conductivity.leading_order_conductivity import (
    LeadingOrder,
)


class BaseLeadingOrderSurfaceForm(LeadingOrder):
    """Base class for leading-order conservation of charge in the electrolyte employing
    the Stefan-Maxwell constitutive equations employing the surface potential difference
    formulation. (Leading refers to leading order in asymptotics)

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain in which the model holds
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, domain, options=None):
        super().__init__(param, domain, options)

    def get_fundamental_variables(self):
        delta_phi_av = pybamm.Variable(
            f"X-averaged {self.domain} electrode surface potential difference [V]",
            domain="current collector",
            reference=self.domain_param.prim.U_init,
        )

        variables = self._get_standard_average_surface_potential_difference_variables(
            delta_phi_av
        )
        variables.update(
            self._get_standard_surface_potential_difference_variables(delta_phi_av)
        )
        return variables

    def get_coupled_variables(self, variables):
        # Only update coupled variables once
        if self.domain == "negative":
            return super().get_coupled_variables(variables)
        else:
            return variables

    def set_initial_conditions(self, variables):
        domain = self.domain
        delta_phi = variables[
            f"X-averaged {domain} electrode surface potential difference [V]"
        ]
        delta_phi_init = self.domain_param.prim.U_init

        self.initial_conditions = {delta_phi: delta_phi_init}

    def set_boundary_conditions(self, variables):
        if self.domain == "negative":
            phi_e = variables["Electrolyte potential [V]"]
            self.boundary_conditions = {
                phi_e: {
                    "left": (pybamm.Scalar(0), "Neumann"),
                    "right": (pybamm.Scalar(0), "Neumann"),
                }
            }


class LeadingOrderDifferential(BaseLeadingOrderSurfaceForm):
    """Leading-order model for conservation of charge in the electrolyte employing the
    Stefan-Maxwell constitutive equations employing the surface potential difference
    formulation and where capacitance is present.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, domain, options=None):
        super().__init__(param, domain, options)

    def set_rhs(self, variables):
        domain = self.domain
        T = variables[f"X-averaged {domain} electrode temperature [K]"]
        C_dl = self.domain_param.C_dl(T)

        delta_phi = variables[
            f"X-averaged {domain} electrode surface potential difference [V]"
        ]

        sum_a_j = variables[
            f"Sum of x-averaged {domain} electrode volumetric "
            "interfacial current densities [A.m-3]"
        ]
        sum_a_j_av = variables[
            f"X-averaged {domain} electrode total volumetric "
            "interfacial current density [A.m-3]"
        ]
        a = variables[
            f"X-averaged {domain} electrode surface area to volume ratio [m-1]"
        ]

        self.rhs[delta_phi] = 1 / (a * C_dl) * (sum_a_j_av - sum_a_j)


class LeadingOrderAlgebraic(BaseLeadingOrderSurfaceForm):
    """Leading-order model for conservation of charge in the electrolyte employing the
    Stefan-Maxwell constitutive equations employing the surface potential difference
    formulation.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, domain, options=None):
        super().__init__(param, domain, options)

    def set_algebraic(self, variables):
        domain = self.domain
        sum_a_j = variables[
            f"Sum of x-averaged {domain} electrode volumetric "
            "interfacial current densities [A.m-3]"
        ]

        sum_a_j_av = variables[
            f"X-averaged {domain} electrode total volumetric "
            "interfacial current density [A.m-3]"
        ]
        delta_phi = variables[
            f"X-averaged {domain} electrode surface potential difference [V]"
        ]

        # multiply by Lx**2 to improve conditioning
        self.algebraic[delta_phi] = (sum_a_j_av - sum_a_j) * self.param.L_x**2



#
# Class for Ordered packing transport_efficiency
#
import pybamm

from .base_transport_efficiency import BaseModel


class OrderedPacking(BaseModel):
    """Submodel for Ordered Packing transport_efficiency
    :footcite:t:`Akanni1987`, :footcite:t:`Shen2007`

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    component : str
        The material for the model ('electrolyte' or 'electrode').
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, component, options=None):
        super().__init__(param, component, options=options)

    def get_coupled_variables(self, variables):
        pybamm.citations.register("Shen2007")
        pybamm.citations.register("Akanni1987")
        if self.component == "Electrolyte":
            tor_dict = {}
            for domain in self.options.whole_cell_domains:
                Domain = domain.capitalize()
                eps_k = variables[f"{Domain} porosity"]
                tor_k = (3 - eps_k) * 0.5
                tor_dict[domain] = tor_k
        elif self.component == "Electrode":
            tor_dict = {}
            for domain in self.options.whole_cell_domains:
                if domain == "separator":
                    tor_k = pybamm.FullBroadcast(0, "separator", "current collector")
                else:
                    Domain = domain.capitalize()
                    phi_k = 1 - variables[f"{Domain} porosity"]
                    tor_k = (3 - phi_k) * 0.5
                tor_dict[domain] = tor_k
        variables.update(self._get_standard_transport_efficiency_variables(tor_dict))

        return variables



#
# Class for no plating
#
import pybamm

from .base_plating import BasePlating


class NoPlating(BasePlating):
    """Base class for no lithium plating/stripping.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, domain, options=None, phase="primary"):
        super().__init__(param, domain, options=options, phase=phase)

    def get_fundamental_variables(self):
        phase_name = self.phase_name
        if self.size_distribution is False:
            zero = pybamm.FullBroadcast(
                pybamm.Scalar(0), f"{self.domain} electrode", "current collector"
            )
        else:
            zero = pybamm.FullBroadcast(
                pybamm.Scalar(0),
                f"{self.domain} {phase_name}particle size",
                {
                    "secondary": f"{self.domain} electrode",
                    "tertiary": "current collector",
                },
            )
        variables = self._get_standard_concentration_variables(zero, zero)
        if self.size_distribution:
            variables.update(
                self._get_standard_size_distribution_overpotential_variables(zero)
            )
            variables.update(
                self._get_standard_size_distribution_reaction_variables(zero)
            )
        else:
            variables.update(self._get_standard_reaction_variables(zero))
        variables.update(self._get_standard_overpotential_variables(zero))

        return variables

    def get_coupled_variables(self, variables):
        # Update whole cell variables, which also updates the "sum of" variables
        variables.update(super().get_coupled_variables(variables))

        return variables



import inspect

import numpy as np

import pybamm


def setup_callbacks(callbacks):
    callbacks = callbacks or []
    if not isinstance(callbacks, list):
        callbacks = [callbacks]

    # Check if there is a logging callback already, if not add the default one
    has_logging_callback = any(isinstance(cb, LoggingCallback) for cb in callbacks)
    if not has_logging_callback:
        callbacks.append(LoggingCallback())

    return CallbackList(callbacks)


class Callback:
    """
    Base class for callbacks, for documenting callback methods.

    Callbacks are used to perform actions (e.g. logging, saving) at certain points in
    the simulation. Each callback method is named `on_<event>`, where `<event>`
    describes the point at which the callback is called. For example, the callback
    `on_experiment_start` is called at the start of an experiment simulation. In
    general, callbacks take a single argument, `logs`, which is a dictionary of
    information about the simulation. Each callback method should return `None`
    (the output of the method is ignored).

    **EXPERIMENTAL** - this class is experimental and the callback interface may
    change in future releases.
    """

    def on_experiment_start(self, logs):
        """
        Called at the start of an experiment simulation.
        """
        pass  # pragma: no cover

    def on_cycle_start(self, logs):
        """
        Called at the start of each cycle in an experiment simulation.
        """
        pass  # pragma: no cover

    def on_step_start(self, logs):
        """
        Called at the start of each step in an experiment simulation.
        """
        pass  # pragma: no cover

    def on_step_end(self, logs):
        """
        Called at the end of each step in an experiment simulation.
        """
        pass  # pragma: no cover

    def on_cycle_end(self, logs):
        """
        Called at the end of each cycle in an experiment simulation.
        """
        pass  # pragma: no cover

    def on_experiment_end(self, logs):
        """
        Called at the end of an experiment simulation.
        """
        pass  # pragma: no cover

    def on_experiment_error(self, logs):
        """
        Called when a SolverError occurs during an experiment simulation.

        For example, this could be used to send an error alert with a bug report when
        running batch simulations in the cloud.
        """
        pass  # pragma: no cover

    def on_experiment_infeasible_time(self, logs):
        """
        Called when an experiment simulation is infeasible due to reaching maximum time.
        """
        pass  # pragma: no cover

    def on_experiment_infeasible_event(self, logs):
        """
        Called when an experiment simulation is infeasible due to an event.
        """
        pass  # pragma: no cover


########################################################################################
class CallbackList(Callback):
    """
    Container abstracting a list of callbacks, so that they can be called in a
    single step e.g. `callbacks.on_simulation_end(...)`.

    This is done without having to redefine the method each time by using the
    `callback_loop_decorator` decorator, which is applied to every method that starts
    with `on_`, using the `inspect` module.

    If better control over how the callbacks are called is required, it might be better
    to be more explicit with the for loop.
    """

    def __init__(self, callbacks):
        self.callbacks = callbacks

    def __len__(self):
        return len(self.callbacks)

    def __getitem__(self, index):
        return self.callbacks[index]


def callback_loop_decorator(func):
    """
    A decorator to call the function on every callback in `self.callbacks`
    """

    def wrapper(self, *args, **kwargs):
        for callback in self.callbacks:
            # call the function on the callback
            getattr(callback, func.__name__)(*args, **kwargs)

    return wrapper


# inspect.getmembers finds all the methods in the Callback class
for name, func in inspect.getmembers(CallbackList, inspect.isfunction):
    if name.startswith("on_"):
        # Replaces each function with the decorated version
        setattr(CallbackList, name, callback_loop_decorator(func))

########################################################################################


class LoggingCallback(Callback):
    """
    Logging callback, implements methods to log progress of the simulation.

    Parameters
    ----------
    logfile : str, optional
        Where to send the log output. If None, uses pybamm's logger.
    """

    def __init__(self, logfile=None):
        self.logfile = logfile
        if logfile is None:
            # Use pybamm's logger, which prints to command line
            self.logger = pybamm.logger
        else:
            # Use a custom logger, this will have its own level so set it to the same
            # level as the pybamm logger (users can override this)
            self.logger = pybamm.get_new_logger(__name__, logfile)
            self.logger.setLevel(pybamm.logger.level)

    def on_experiment_start(self, logs):
        # Clear the log file
        self.logger.info("Start running experiment")
        if self.logfile is not None:
            with open(self.logfile, "w") as f:
                f.write("")

    def on_cycle_start(self, logs):
        cycle_num, num_cycles = logs["cycle number"]
        total_time = logs["elapsed time"]
        self.logger.notice(
            f"Cycle {cycle_num}/{num_cycles} ({total_time} elapsed) " + "-" * 20
        )

    def on_step_start(self, logs):
        cycle_num, num_cycles = logs["cycle number"]
        step_num, cycle_length = logs["step number"]
        operating_conditions = logs["step operating conditions"]
        self.logger.notice(
            f"Cycle {cycle_num}/{num_cycles}, step {step_num}/{cycle_length}: "
            f"{operating_conditions}"
        )

    def on_step_end(self, logs):
        time_stop = logs["stopping conditions"]["time"]
        if time_stop is not None:
            time_now = logs["experiment time"]
            if time_now < time_stop:
                self.logger.notice(
                    f"Time is now {time_now:.3f} s, will stop at {time_stop:.3f} s."
                )
            else:
                self.logger.notice(
                    f"Stopping experiment since time ({time_now:.3f} s) "
                    f"has reached stopping time ({time_stop:.3f} s)."
                )

    def on_cycle_end(self, logs):
        cap_stop = logs["stopping conditions"].get("capacity")
        if cap_stop is not None:
            cap_now = logs["summary variables"]["Capacity [A.h]"]
            cap_start = logs["start capacity"]
            if np.isnan(cap_now) or cap_now > cap_stop:
                self.logger.notice(
                    f"Capacity is now {cap_now:.3f} Ah (originally {cap_start:.3f} Ah, "
                    f"will stop at {cap_stop:.3f} Ah)"
                )
            else:
                self.logger.notice(
                    f"Stopping experiment since capacity ({cap_now:.3f} Ah) "
                    f"is below stopping capacity ({cap_stop:.3f} Ah)."
                )

        voltage_stop = logs["stopping conditions"]["voltage"]
        if voltage_stop is not None:
            min_voltage = logs["Minimum voltage [V]"]
            if min_voltage > voltage_stop[0]:
                self.logger.notice(
                    f"Minimum voltage is now {min_voltage:.3f} V "
                    f"(will stop at {voltage_stop[0]:.3f} V)"
                )
            else:
                self.logger.notice(
                    f"Stopping experiment since minimum voltage ({min_voltage:.3f} V) "
                    f"is below stopping voltage ({voltage_stop[0]:.3f} V)."
                )

    def on_experiment_end(self, logs):
        elapsed_time = logs["elapsed time"]
        self.logger.notice(f"Finish experiment simulation, took {elapsed_time}")

    def on_experiment_error(self, logs):
        error = logs["error"]
        pybamm.logger.error(f"Simulation error: {error}")

    def on_experiment_infeasible_time(self, logs):
        duration = logs["step duration"]
        cycle_num = logs["cycle number"][0]
        step_num = logs["step number"][0]
        operating_conditions = logs["step operating conditions"]
        self.logger.warning(
            f"\n\n\tExperiment is infeasible: default duration ({duration} seconds) "
            f"was reached during '{operating_conditions}'. The returned solution only "
            f"contains up to step {step_num} of cycle {cycle_num}. "
            "Please specify a duration in the step instructions."
        )

    def on_experiment_infeasible_event(self, logs):
        termination = logs["termination"]
        cycle_num = logs["cycle number"][0]
        step_num = logs["step number"][0]
        operating_conditions = logs["step operating conditions"]
        self.logger.warning(
            f"\n\n\tExperiment is infeasible: '{termination}' was "
            f"triggered during '{operating_conditions}'. The returned solution only "
            f"contains up to step {step_num} of cycle {cycle_num}. "
        )



#
# Base class for convection submodels in the through-cell direction
#
import pybamm
from pybamm.models.submodels.convection.base_convection import BaseModel


class BaseThroughCellModel(BaseModel):
    """Base class for convection submodels in the through-cell direction.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict, optional
        A dictionary of options to be passed to the model.
    """

    def __init__(self, param, options=None):
        super().__init__(param, options=options)

    def _get_standard_convection_variables(self, domain, v_box_k, div_v_box_k, p_k):
        Domain = domain.capitalize()

        div_v_box_k_av = pybamm.x_average(div_v_box_k)

        variables = {
            f"{Domain} volume-averaged velocity [m.s-1]": v_box_k,
            f"{Domain} volume-averaged acceleration [m.s-2]": div_v_box_k,
            f"X-averaged {domain} volume-averaged acceleration [m.s-2]": div_v_box_k_av,
            f"{Domain} pressure [Pa]": p_k,
            f"X-averaged {domain} pressure [Pa]": pybamm.x_average(p_k),
        }
        return variables

    def _get_standard_sep_velocity_variables(self, v_box_s, div_v_box_s):
        """Volume-averaged velocity in the separator"""

        div_v_box_s_av = pybamm.x_average(div_v_box_s)

        variables = {
            "Separator volume-averaged velocity [m.s-1]": v_box_s,
            "Separator volume-averaged acceleration [m.s-2]": div_v_box_s,
            "X-averaged separator volume-averaged acceleration [m.s-2]": div_v_box_s_av,
        }

        return variables



#
# Base interface class
#

import pybamm


class BaseInterface(pybamm.BaseSubModel):
    """
    Base class for interfacial currents

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain to implement the model, either: 'Negative' or 'Positive'.
    reaction : str
        The name of the reaction being implemented
    options: dict
        A dictionary of options to be passed to the model. See
        :class:`pybamm.BaseBatteryModel`
    phase : str, optional
        Phase of the particle (default is "primary")
    """

    def __init__(self, param, domain, reaction, options, phase="primary"):
        super().__init__(param, domain, options=options, phase=phase)
        if reaction in ["lithium-ion main", "lithium metal plating"]:
            self.reaction_name = ""
        elif reaction == "lead-acid main":
            self.reaction_name = ""  # empty reaction name for the main reaction
        elif reaction == "lead-acid oxygen":
            self.reaction_name = "oxygen "
        elif reaction in ["SEI", "SEI on cracks", "lithium plating"]:
            self.reaction_name = reaction + " "

        if reaction in [
            "lithium-ion main",
            "lithium metal plating",
            "lithium plating",
            "SEI",
            "SEI on cracks",
        ]:
            # phase_name can be "" or "primary " or "secondary "
            self.reaction_name = self.phase_name + self.reaction_name

        self.reaction = reaction
        domain_options = getattr(self.options, domain)
        self.size_distribution = domain_options["particle size"] == "distribution"

    def _get_exchange_current_density(self, variables):
        """
        A private function to obtain the exchange current density

        Parameters
        ----------
        variables: dict
            The variables in the full model.

        Returns
        -------
        j0 : :class: `pybamm.Symbol`
            The exchange current density.
        """
        phase_param = self.phase_param
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name
        domain_options = getattr(self.options, domain)

        c_e = variables[f"{Domain} electrolyte concentration [mol.m-3]"]
        T = variables[f"{Domain} electrode temperature [K]"]

        if self.reaction == "lithium-ion main":
            # For "particle-size distribution" submodels, take distribution version
            # of c_s_surf that depends on particle size.
            domain_options = getattr(self.options, domain)
            if domain_options["particle size"] == "distribution":
                c_s_surf = variables[
                    f"{Domain} {phase_name}particle surface "
                    "concentration distribution [mol.m-3]"
                ]
                # If all variables were broadcast (in "x"), take only the orphans,
                # then re-broadcast c_e
                if (
                    isinstance(c_s_surf, pybamm.Broadcast)
                    and isinstance(c_e, pybamm.Broadcast)
                    and isinstance(T, pybamm.Broadcast)
                ):
                    c_s_surf = c_s_surf.orphans[0]
                    c_e = c_e.orphans[0]
                    T = T.orphans[0]

                    # as c_e must now be a scalar, re-broadcast to
                    # "current collector"
                    c_e = pybamm.PrimaryBroadcast(c_e, ["current collector"])
                # broadcast c_e, T onto "particle size"
                c_e = pybamm.PrimaryBroadcast(
                    c_e, [f"{domain} {phase_name}particle size"]
                )
                T = pybamm.PrimaryBroadcast(T, [f"{domain} {phase_name}particle size"])

            else:
                c_s_surf = variables[
                    f"{Domain} {phase_name}particle surface concentration [mol.m-3]"
                ]
                # If all variables were broadcast, take only the orphans
                if (
                    isinstance(c_s_surf, pybamm.Broadcast)
                    and isinstance(c_e, pybamm.Broadcast)
                    and isinstance(T, pybamm.Broadcast)
                ):
                    c_s_surf = c_s_surf.orphans[0]
                    c_e = c_e.orphans[0]
                    T = T.orphans[0]
            # Get main reaction exchange-current density (may have empirical hysteresis)
            j0_option = getattr(domain_options, self.phase)["exchange-current density"]
            if j0_option == "single":
                j0 = phase_param.j0(c_e, c_s_surf, T)
            elif j0_option == "current sigmoid":
                current = variables["Total current density [A.m-2]"]
                k = 100
                if Domain == "Positive":
                    lithiation_current = current
                elif Domain == "Negative":
                    lithiation_current = -current
                m_lith = pybamm.sigmoid(
                    0, lithiation_current, k
                )  # lithiation_current > 0
                m_delith = 1 - m_lith  # lithiation_current < 0
                j0_lith = phase_param.j0(c_e, c_s_surf, T, "lithiation")
                j0_delith = phase_param.j0(c_e, c_s_surf, T, "delithiation")
                j0 = m_lith * j0_lith + m_delith * j0_delith

        elif self.reaction == "lithium metal plating":
            # compute T on the surface of the anode (interface with separator)
            T = pybamm.boundary_value(T, "right")
            c_Li_metal = 1 / self.param.V_bar_Li
            j0 = self.param.j0_Li_metal(c_e, c_Li_metal, T)

        elif self.reaction == "lead-acid main":
            # If variable was broadcast, take only the orphan
            if isinstance(c_e, pybamm.Broadcast) and isinstance(T, pybamm.Broadcast):
                c_e = c_e.orphans[0]
                T = T.orphans[0]
            j0 = phase_param.j0(c_e, T)

        elif self.reaction == "lead-acid oxygen":
            # If variable was broadcast, take only the orphan
            if isinstance(c_e, pybamm.Broadcast) and isinstance(T, pybamm.Broadcast):
                c_e = c_e.orphans[0]
                T = T.orphans[0]
            if self.domain == "negative":
                j0 = pybamm.Scalar(0)
            elif self.domain == "positive":
                j0 = self.param.p.prim.j0_Ox(c_e, T)

        return j0

    def _get_number_of_electrons_in_reaction(self):
        """Returns the number of electrons in the reaction."""
        if self.reaction in [
            "lithium-ion main",
            "lithium metal plating",
        ]:
            return self.phase_param.ne
        elif self.reaction == "lead-acid main":
            return self.phase_param.ne
        elif self.reaction == "lead-acid oxygen":
            return self.param.ne_Ox

    def _get_average_total_interfacial_current_density(self, variables):
        """
        Method to obtain the average total interfacial current density.

        Note: for lithium-ion models this is only exact if all the particles have
        the same radius. For the current set of models implemeted in pybamm,
        having the radius as a function of through-cell distance only makes sense
        for the DFN model. In the DFN, the correct average interfacial current density
        is computed in 'base_kinetics.py' by averaging the actual interfacial current
        density. The approximation here is only used to get the approximate constant
        additional resistance term for the "average" SEI film resistance model
        (if using), where only negligible errors will be introduced.

        For "leading-order" and "composite" submodels (as used in the SPM and SPMe)
        there is only a single particle radius, so this method returns correct result.
        """
        domain = self.domain
        i_boundary_cc = variables["Current collector current density [A.m-2]"]

        if self.options.electrode_types[domain] == "planar":
            # In a half-cell the total interfacial current density is the current
            # collector current density, not divided by electrode thickness
            i_boundary_cc = variables["Current collector current density [A.m-2]"]
            j_total_average = i_boundary_cc
            a_j_total_average = i_boundary_cc
        else:
            a_av = variables[
                f"X-averaged {domain} electrode {self.phase_name}"
                "surface area to volume ratio [m-1]"
            ]
            sgn = 1 if self.domain == "negative" else -1

            a_j_total_average = sgn * i_boundary_cc / (self.domain_param.L)
            j_total_average = a_j_total_average / a_av

        return j_total_average, a_j_total_average

    def _get_standard_interfacial_current_variables(self, j):
        domain, Domain = self.domain_Domain
        reaction_name = self.reaction_name

        j.print_name = f"j_{domain[0]}"

        if self.reaction == "lithium metal plating":
            # Half-cell domain, j should not be broadcast
            variables = {"Lithium metal plating current density [A.m-2]": j}
            return variables

        # Size average. For j variables that depend on particle size, see
        # "_get_standard_size_distribution_interfacial_current_variables"
        if j.domain in [
            ["negative particle size"],
            ["positive particle size"],
            ["negative primary particle size"],
            ["positive primary particle size"],
            ["negative secondary particle size"],
            ["positive secondary particle size"],
        ]:
            j = pybamm.size_average(j)
        # Average, and broadcast if necessary
        j_av = pybamm.x_average(j)
        if j.domain == []:
            j = pybamm.FullBroadcast(j, f"{domain} electrode", "current collector")
        elif (
            j.domain == ["current collector"]
            and getattr(self, "reaction_loc", None) != "interface"
        ):
            j = pybamm.PrimaryBroadcast(j, f"{domain} electrode")

        variables = {
            f"{Domain} electrode {reaction_name}interfacial current density [A.m-2]": j,
            f"X-averaged {domain} electrode {reaction_name}"
            "interfacial current density [A.m-2]": j_av,
        }
        if self.phase_name == reaction_name:
            variables.update(
                {
                    f"{Domain} electrode {reaction_name}interfacial current density [A.m-2]": j,
                    f"X-averaged {domain} electrode {reaction_name}interfacial current density [A.m-2]": j_av,
                }
            )

        return variables

    def _get_standard_total_interfacial_current_variables(self, j_tot_av, a_j_tot_av):
        domain = self.domain

        if self.options.electrode_types[domain] == "planar":
            variables = {
                "Lithium metal total interfacial current density [A.m-2]": j_tot_av,
            }
        else:
            variables = {
                f"X-averaged {domain} electrode total interfacial "
                "current density [A.m-2]": j_tot_av,
                f"X-averaged {domain} electrode total volumetric interfacial "
                "current density [A.m-3]": a_j_tot_av,
            }

        return variables

    def _get_standard_exchange_current_variables(self, j0):
        domain, Domain = self.domain_Domain
        reaction_name = self.reaction_name

        if self.reaction == "lithium metal plating":
            # half-cell domain
            variables = {
                "Lithium metal interface exchange current density [A.m-2]": j0,
            }
            return variables

        # Size average. For j0 variables that depend on particle size, see
        # "_get_standard_size_distribution_exchange_current_variables"
        if j0.domain in [["negative particle size"], ["positive particle size"]]:
            j0 = pybamm.size_average(j0)
        # Average, and broadcast if necessary
        j0_av = pybamm.x_average(j0)

        # X-average, and broadcast if necessary
        if j0.domain == []:
            j0 = pybamm.FullBroadcast(j0, f"{domain} electrode", "current collector")
        elif j0.domain == ["current collector"]:
            j0 = pybamm.PrimaryBroadcast(j0, f"{domain} electrode")

        variables = {
            f"{Domain} electrode {reaction_name}exchange current density [A.m-2]": j0,
            f"X-averaged {domain} electrode {reaction_name}"
            "exchange current density [A.m-2]": j0_av,
        }

        return variables

    def _get_standard_volumetric_current_density_variables(self, variables):
        domain, Domain = self.domain_Domain

        if self.options.electrode_types[domain] == "planar":
            return variables

        reaction_name = self.reaction_name
        phase_name = self.phase_name

        if isinstance(self, pybamm.kinetics.NoReaction):
            a = 0
        else:
            a = variables[
                f"{Domain} electrode {phase_name}surface area to volume ratio [m-1]"
            ]
        j = variables[
            f"{Domain} electrode {reaction_name}interfacial current density [A.m-2]"
        ]
        a_j = a * j
        a_j_av = pybamm.x_average(a_j)

        if reaction_name == "SEI on cracks ":
            roughness = variables[f"{Domain} {phase_name}electrode roughness ratio"] - 1
            roughness_av = (
                variables[f"X-averaged {domain} {phase_name}electrode roughness ratio"]
                - 1
            )
        else:
            roughness = 1
            roughness_av = 1

        variables.update(
            {
                f"{Domain} electrode {reaction_name}volumetric "
                "interfacial current density [A.m-3]": a_j * roughness,
                f"X-averaged {domain} electrode {reaction_name}volumetric interfacial "
                "current density [A.m-3]": a_j_av * roughness_av,
            }
        )
        return variables

    def _get_standard_overpotential_variables(self, eta_r):
        domain, Domain = self.domain_Domain
        reaction_name = self.reaction_name

        if self.reaction == "lithium metal plating":
            # half-cell domain
            variables = {"Lithium metal interface reaction overpotential [V]": eta_r}
            return variables

        # Size average. For eta_r variables that depend on particle size, see
        # "_get_standard_size_distribution_overpotential_variables"
        if eta_r.domain in [
            ["negative particle size"],
            ["positive particle size"],
            ["negative primary particle size"],
            ["positive primary particle size"],
            ["negative secondary particle size"],
            ["positive secondary particle size"],
        ]:
            eta_r = pybamm.size_average(eta_r)

        # X-average, and broadcast if necessary
        eta_r_av = pybamm.x_average(eta_r)
        if eta_r.domain == ["current collector"]:
            eta_r = pybamm.PrimaryBroadcast(eta_r, f"{domain} electrode")

        variables = {
            f"{Domain} electrode {reaction_name}reaction overpotential [V]": eta_r,
            f"X-averaged {domain} electrode {reaction_name}reaction "
            "overpotential [V]": eta_r_av,
        }

        return variables

    def _get_standard_sei_film_overpotential_variables(self, eta_sei):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name
        Phase_name = phase_name.capitalize()

        if self.options.electrode_types[domain] == "planar":
            # half-cell domain
            variables = {
                f"{Domain} electrode {Phase_name}SEI film overpotential [V]": eta_sei,
            }
            return variables

        # Average, and broadcast if necessary
        eta_sei_av = pybamm.x_average(eta_sei)
        if eta_sei.domain == []:
            eta_sei = pybamm.FullBroadcast(
                eta_sei, f"{domain} electrode", "current collector"
            )
        elif eta_sei.domain == ["current collector"]:
            eta_sei = pybamm.PrimaryBroadcast(eta_sei, f"{domain} electrode")

        variables = {
            f"{Domain} electrode {phase_name}SEI film overpotential [V]": eta_sei,
            f"X-averaged {domain} electrode {phase_name}SEI"
            " film overpotential [V]": eta_sei_av,
        }

        return variables

    def _get_standard_average_surface_potential_difference_variables(
        self, delta_phi_av
    ):
        domain = self.domain

        if self.options.electrode_types[domain] == "planar":
            variables = {
                "Lithium metal interface surface potential difference [V]"
                "": delta_phi_av,
            }
        else:
            variables = {
                f"X-averaged {domain} electrode "
                "surface potential difference [V]": delta_phi_av,
            }

        return variables

    def _get_standard_size_distribution_interfacial_current_variables(self, j):
        """
        Interfacial current density variables that depend on particle size R,
        relevant if "particle size" option is "distribution".
        """
        domain, Domain = self.domain_Domain
        reaction_name = self.reaction_name

        # X-average and broadcast if necessary
        if j.domains["secondary"] == [f"{domain} electrode"]:
            # x-average
            j_xav = pybamm.x_average(j)
        elif getattr(self, "reaction_loc", None) != "interface":
            j_xav = j
            j = pybamm.SecondaryBroadcast(j_xav, [f"{domain} electrode"])
        else:
            j_xav = j

        variables = {
            f"{Domain} electrode {reaction_name}"
            "interfacial current density distribution [A.m-2]": j,
            f"X-averaged {domain} electrode {reaction_name}"
            "interfacial current density distribution [A.m-2]": j_xav,
        }

        return variables

    def _get_standard_size_distribution_exchange_current_variables(self, j0):
        """
        Exchange current variables that depend on particle size.
        """
        domain, Domain = self.domain_Domain
        reaction_name = self.reaction_name

        # X-average or broadcast to electrode if necessary
        if j0.domains["secondary"] != [f"{domain} electrode"]:
            j0_av = j0
            j0 = pybamm.SecondaryBroadcast(j0, f"{domain} electrode")
        else:
            j0_av = pybamm.x_average(j0)

        variables = {
            f"{Domain} electrode {reaction_name}"
            "exchange current density distribution [A.m-2]": j0,
            f"X-averaged {domain} electrode {reaction_name}"
            "exchange current density distribution [A.m-2]": j0_av,
        }

        return variables

    def _get_standard_size_distribution_overpotential_variables(self, eta_r):
        """
        Overpotential variables that depend on particle size.
        """
        domain, Domain = self.domain_Domain
        reaction_name = self.reaction_name

        # X-average or broadcast to electrode if necessary
        if eta_r.domains["secondary"] != [f"{domain} electrode"]:
            eta_r_av = eta_r
            eta_r = pybamm.SecondaryBroadcast(eta_r, f"{domain} electrode")
        else:
            eta_r_av = pybamm.x_average(eta_r)

        domain_reaction = f"{Domain} electrode {reaction_name}reaction overpotential"

        variables = {
            f"{domain_reaction} [V]": eta_r,
            f"X-averaged {domain_reaction.lower()} distribution [V]": eta_r_av,
        }

        return variables



#
# Prettify print_name
#
PRINT_NAME_OVERRIDES = {
    "current_with_time": "I",
    "current_density_with_time": r"i_{\mathrm{cell}}",
    "thermodynamic_factor": r"\left(1+\frac{dlnf}{dlnc}\right)",
    "t_plus": r"t_{\mathrm{+}}",
}

GREEK_LETTERS = [
    "alpha",
    "beta",
    "gamma",
    "delta",
    "epsilon",
    "zeta",
    "eta",
    "theta",
    "iota",
    "kappa",
    "lambda",
    "mu",
    "nu",
    "xi",
    "pi",
    "rho",
    "sigma",
    "tau",
    "upsilon",
    "phi",
    "chi",
    "psi",
    "omega",
]


def prettify_print_name(name):
    """Prettify print_name using regex"""

    # Skip prettify_print_name() for cases like `new_copy()`
    if name is None or "{" in name or "\\" in name:
        return name

    # Find subscripts, superscripts, and averaging
    # Remove them from the name one by one and add them later in processed form
    subscripts = []
    superscripts = []
    average = False

    processing = True
    while processing:
        # Set processing to False. If any of the following conditions are met,
        # it will be set to True again
        processing = False
        for superscript in ["init", "ref", "typ", "max", "0", "surf"]:
            if f"_{superscript}_" in name or name.endswith(f"_{superscript}"):
                superscripts.append(superscript)
                name = name.replace(f"_{superscript}", "")
                processing = True
                break
        # "0" might also appear without a preceding underscore
        for superscript in ["0"]:
            if superscript in name:
                superscripts.append(superscript)
                name = name.replace(superscript, "")
                processing = True
                break
        for subscript in ["cc", "dl", "R", "e", "s", "n", "p", "amb"]:
            if f"_{subscript}_" in name or name.endswith(f"_{subscript}"):
                subscripts.append(subscript)
                name = name.replace(f"_{subscript}", "")
                processing = True
                break
        for av in ["av", "xav"]:
            if f"_{av}_" in name or name.endswith(f"_{av}"):
                average = True
                name = name.replace(f"_{av}", "")
                processing = True
                break

    # Process name
    # Override print_name if name exists in the dictionary
    if name in PRINT_NAME_OVERRIDES:
        name = PRINT_NAME_OVERRIDES[name]

    # Replace eps with epsilon (e.g. eps_n --> epsilon_n)
    if name == "eps":
        name = r"\epsilon"
    if name == "eps_c":
        name = r"(\epsilon c)"

    # Greek letters (delta --> \delta)
    if name.lower() in GREEK_LETTERS:
        name = "\\" + name

    # Add subscripts and superscripts
    if average:
        name = r"\overline{" + name + "}"
    if subscripts:
        name += r"_{\mathrm{" + ",".join(subscripts) + "}}"
    if superscripts:
        name += r"^{\mathrm{" + ",".join(superscripts) + "}}"

    return name



#
# Open-circuit potential from the Multi-Species Multi-Reaction framework
#
import pybamm

from . import BaseOpenCircuitPotential


class MSMROpenCircuitPotential(BaseOpenCircuitPotential):
    """
    Class for open-circuit potential within the Multi-Species Multi-Reaction
    framework :footcite:t:`Baker2018`. The thermodynamic model is presented in
    :footcite:t:`Verbrugge2017`, along with parameter values for a number of
    substitutional materials.
    """

    def __init__(
        self, param, domain, reaction, options, phase="primary", x_average=False
    ):
        super().__init__(
            param, domain, reaction, options=options, phase=phase, x_average=x_average
        )
        pybamm.citations.register("Baker2018")

    def get_coupled_variables(self, variables):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        if self.reaction == "lithium-ion main":
            T = variables[f"{Domain} electrode temperature [K]"]
            # For "particle-size distribution" models, take distribution version
            # of c_s_surf that depends on particle size.
            domain_options = getattr(self.options, domain)
            if domain_options["particle size"] == "distribution":
                sto_surf = variables[
                    f"{Domain} {phase_name}particle surface stoichiometry distribution"
                ]
                ocp_surf = variables[
                    f"{Domain} {phase_name}particle surface potential distribution [V]"
                ]
                # If variable was broadcast, take only the orphan
                if (
                    isinstance(sto_surf, pybamm.Broadcast)
                    and isinstance(ocp_surf, pybamm.Broadcast)
                    and isinstance(T, pybamm.Broadcast)
                ):
                    sto_surf = sto_surf.orphans[0]
                    ocp_surf = ocp_surf.orphans[0]
                    T = T.orphans[0]
                T = pybamm.PrimaryBroadcast(T, [f"{domain} particle size"])
            else:
                sto_surf = variables[
                    f"{Domain} {phase_name}particle surface stoichiometry"
                ]
                ocp_surf = variables[
                    f"{Domain} {phase_name}particle surface potential [V]"
                ]
                # If variable was broadcast, take only the orphan
                if (
                    isinstance(sto_surf, pybamm.Broadcast)
                    and isinstance(ocp_surf, pybamm.Broadcast)
                    and isinstance(T, pybamm.Broadcast)
                ):
                    sto_surf = sto_surf.orphans[0]
                    ocp_surf = ocp_surf.orphans[0]
                    T = T.orphans[0]

            ocp_bulk = variables[f"Average {domain} {phase_name}particle potential [V]"]
            dUdT = self.phase_param.dUdT(sto_surf)

        variables.update(self._get_standard_ocp_variables(ocp_surf, ocp_bulk, dUdT))
        variables.update(
            {
                f"{Domain} electrode {phase_name}equilibrium open-circuit potential [V]": ocp_surf,
                f"X-averaged {domain} electrode {phase_name}equilibrium open-circuit potential [V]": pybamm.x_average(
                    ocp_surf
                ),
            }
        )
        return variables



from .base_transport_efficiency import BaseModel
from .bruggeman import Bruggeman
from .cation_exchange_membrane import CationExchangeMembrane
from .heterogeneous_catalyst import HeterogeneousCatalyst
from .hyperbola_of_revolution import HyperbolaOfRevolution
from .ordered_packing import OrderedPacking
from .overlapping_spheres import OverlappingSpheres
from .random_overlapping_cylinders import RandomOverlappingCylinders
from .tortuosity_factor import TortuosityFactor

__all__ = ['base_transport_efficiency', 'bruggeman', 'cation_exchange_membrane',
           'heterogeneous_catalyst', 'hyperbola_of_revolution', 'ordered_packing', 'overlapping_spheres',
           'random_overlapping_cylinders', 'tortuosity_factor']



import pybamm
from pybamm.util import import_optional_dependency


def plot_3d_heatmap(
    solution: "pybamm.Solution",
    variable: str,
    t: float | None,
    ax=None,
    show_plot: bool = True,
    cmap: str = "inferno",
    marker_size: float = 10,
    alpha: float = 0.7,
    use_offset: bool = False,
    **kwargs,
):
    """
    Creates a 3D scatter plot of a variable from a PyBaMM solution with a 3D mesh.

    This function visualizes the solution directly on the FEM nodes, providing a true
    representation of the 3D data.

    Parameters
    ----------
    solution : pybamm.Solution
        The solution object containing the 3D variable.
    variable : str, optional
        The name of the 3D variable to plot (default: "Cell temperature [K]").
    t : float, optional
        The time at which to plot. If None, the last timestep is used.
    ax : matplotlib.axes.Axes, optional
        A 3D axes object on which to draw the plot. If None, a new figure
        and axes are created.
    show_plot : bool, optional
        Whether to display the plot (default: True).
    cmap : str, optional
        The colormap for the plot (default: 'inferno').
    marker_size : float, optional
        The size of the markers in the scatter plot (default: 10).
    alpha : float, optional
        The transparency of the markers (0=transparent, 1=opaque). Default is 0.7.
    use_offset : bool, optional
        If True, uses scientific notation for the color bar (default: False).
    **kwargs
        Additional keyword arguments passed to matplotlib.axes.Axes.scatter.
    """
    plt = import_optional_dependency("matplotlib.pyplot")
    model = solution.all_models[0]
    if model.options.get("dimensionality") != 3:
        raise TypeError("This function requires a 3D model solution.")

    if t is None:
        t = solution.t[-1]

    var_obj = solution[variable]
    mesh = var_obj.mesh
    nodes = mesh.nodes
    color_data = var_obj(t=t, x=nodes[:, 0], y=nodes[:, 1], z=nodes[:, 2])

    if ax is None:
        fig = plt.figure(figsize=(10, 8))
        ax = fig.add_subplot(projection="3d")
    else:
        if ax.name != "3d":
            raise TypeError("The provided axes `ax` must be a 3D projection.")
        fig = ax.get_figure()

    scatter = ax.scatter(
        nodes[:, 0],
        nodes[:, 1],
        nodes[:, 2],
        c=color_data,
        cmap=cmap,
        s=marker_size,
        alpha=alpha,
        **kwargs,
    )

    ax.set_xlabel("x [m]")
    ax.set_ylabel("y [m]")
    ax.set_zlabel("z [m]")
    ax.set_title(f"3D Heatmap of {variable}\nat t={t:.1f}s")

    cbar = fig.colorbar(scatter, ax=ax, shrink=0.6, label=variable)
    if not use_offset:
        cbar.formatter.set_useOffset(False)
        fig.canvas.draw()

    ax.view_init(elev=20, azim=-65)
    fig.tight_layout()

    if show_plot:
        plt.show()

    return ax



from .base_oxygen_diffusion import BaseModel
from .leading_oxygen_diffusion import LeadingOrder
from .full_oxygen_diffusion import Full
from .no_oxygen import NoOxygen

__all__ = ['base_oxygen_diffusion', 'full_oxygen_diffusion',
           'leading_oxygen_diffusion', 'no_oxygen']



#
# Base kinetics class
#
import pybamm
from pybamm.models.submodels.interface.base_interface import BaseInterface


class BaseKinetics(BaseInterface):
    """
    Base submodel for kinetics

    Parameters
    ----------
    param :
        model parameters
    domain : str
        The domain to implement the model, either: 'Negative' or 'Positive'.
    reaction : str
        The name of the reaction being implemented
    options: dict
        A dictionary of options to be passed to the model.
        See :class:`pybamm.BaseBatteryModel`
    phase : str, optional
        Phase of the particle (default is "primary")
    """

    def __init__(self, param, domain, reaction, options, phase="primary"):
        super().__init__(param, domain, reaction, options=options, phase=phase)

    def get_fundamental_variables(self):
        domain = self.domain
        phase_name = self.phase_name

        if (
            self.options["total interfacial current density as a state"] == "true"
            and "main" in self.reaction
        ):
            j = pybamm.Variable(
                f"Total {domain} electrode {phase_name}"
                "interfacial current density variable [A.m-2]",
                domain=f"{domain} electrode",
                auxiliary_domains={"secondary": "current collector"},
            )

            variables = {
                f"Total {domain} electrode {phase_name}"
                "interfacial current density variable [A.m-2]": j,
                f"X-averaged total {domain} electrode {phase_name}"
                "interfacial current density variable [A.m-2]": pybamm.x_average(j),
            }
            return variables
        else:
            return {}

    def get_coupled_variables(self, variables):
        domain, Domain = self.domain_Domain
        reaction_name = self.reaction_name
        phase_name = self.phase_name

        # Get surface potential difference
        if self.reaction == "lithium metal plating":  # li metal electrode (half-cell)
            delta_phi = variables[
                "Lithium metal interface surface potential difference [V]"
            ]
        else:
            delta_phi = variables[
                f"{Domain} electrode surface potential difference [V]"
            ]
            # If delta_phi was broadcast, take only the orphan.
            if isinstance(delta_phi, pybamm.Broadcast):
                delta_phi = delta_phi.orphans[0]
        # For "particle-size distribution" models, delta_phi must then be
        # broadcast to "particle size" domain
        domain_options = getattr(self.options, domain)
        if (
            self.reaction == "lithium-ion main"
            and domain_options["particle size"] == "distribution"
        ):
            delta_phi = pybamm.PrimaryBroadcast(
                delta_phi, [f"{domain} {phase_name}particle size"]
            )

        # Get exchange-current density. For MSMR models we calculate the exchange
        # current density for each reaction, then sum these to give a total exchange
        # current density. Note: this is only used for the "exchange current density"
        # variables. For the interfacial current density variables, we sum the
        # interfacial currents from each reaction.
        if domain_options["intercalation kinetics"] == "MSMR":
            N = int(domain_options["number of MSMR reactions"])
            j0 = 0
            for i in range(N):
                j0_j = self._get_exchange_current_density_by_reaction(variables, i)
                variables.update(
                    self._get_standard_exchange_current_by_reaction_variables(j0_j, i)
                )
                j0 += j0_j
        else:
            j0 = self._get_exchange_current_density(variables)

        # Get open-circuit potential variables and reaction overpotential
        if (
            domain_options["particle size"] == "distribution"
            and self.options.electrode_types[domain] == "porous"
        ):
            ocp = variables[
                f"{Domain} electrode {reaction_name}"
                "open-circuit potential distribution [V]"
            ]
        else:
            ocp = variables[
                f"{Domain} electrode {reaction_name}open-circuit potential [V]"
            ]
        # If ocp was broadcast, and the reaction is lithium metal plating OR
        # delta_phi's secondary domain is "current collector", then take only the
        # orphan.
        if isinstance(ocp, pybamm.Broadcast):
            if self.reaction == "lithium metal plating":
                ocp = ocp.orphans[0]
            elif delta_phi.domains["secondary"] == ["current collector"]:
                ocp = ocp.orphans[0]

        # Get reaction overpotential
        eta_r = delta_phi - ocp

        # Get average interfacial current density
        j_tot_av, a_j_tot_av = self._get_average_total_interfacial_current_density(
            variables
        )
        # Add SEI resistance
        if self.options.electrode_types[domain] == "planar":
            R_sei = self.phase_param.R_sei
            L_sei = variables[f"{Domain} {phase_name}SEI thickness [m]"]  # on interface
            eta_sei = -j_tot_av * L_sei * R_sei
        elif self.options["SEI film resistance"] == "average":
            R_sei = self.phase_param.R_sei
            L_sei_av = variables[f"X-averaged {domain} {phase_name}SEI thickness [m]"]
            eta_sei = -j_tot_av * L_sei_av * R_sei
        elif self.options["SEI film resistance"] == "distributed":
            R_sei = self.phase_param.R_sei
            L_sei = variables[f"{Domain} {phase_name}SEI thickness [m]"]
            j_tot = variables[
                f"Total {domain} electrode {phase_name}"
                "interfacial current density variable [A.m-2]"
            ]

            # Override print_name
            j_tot.print_name = "j_tot"

            eta_sei = -j_tot * L_sei * R_sei
            if self.size_distribution and eta_r.domains["secondary"] != [
                f"{domain} electrode"
            ]:
                eta_sei = pybamm.x_average(eta_sei)
        else:
            eta_sei = pybamm.Scalar(0)
        eta_r += eta_sei

        # Broadcast j0 to match eta_r's domain, if necessary
        if j0.secondary_domain == ["current collector"] and eta_r.secondary_domain == [
            f"{domain} electrode"
        ]:
            j0 = pybamm.SecondaryBroadcast(j0, [f"{domain} electrode"])

        # Get number of electrons in reaction
        ne = self._get_number_of_electrons_in_reaction()

        # Get kinetics. Note: T and u must have the same domain as j0 and eta_r
        if self.options.electrode_types[domain] == "planar":
            T = variables["X-averaged cell temperature [K]"]
            u = variables["Lithium metal interface utilisation"]
        elif j0.domain in ["current collector", ["current collector"]]:
            T = variables["X-averaged cell temperature [K]"]
            u = variables[f"X-averaged {domain} electrode interface utilisation"]
        elif j0.domain == [f"{domain} {phase_name}particle size"]:
            if j0.domains["secondary"] != [f"{domain} electrode"]:
                T = variables["X-averaged cell temperature [K]"]
                u = variables[f"X-averaged {domain} electrode interface utilisation"]
            else:
                T = variables[f"{Domain} electrode temperature [K]"]
                u = variables[f"{Domain} electrode interface utilisation"]

            # Broadcast T onto "particle size" domain
            T = pybamm.PrimaryBroadcast(T, [f"{domain} {phase_name}particle size"])
        else:
            T = variables[f"{Domain} electrode temperature [K]"]
            u = variables[f"{Domain} electrode interface utilisation"]

        # Update j, except in the "distributed SEI resistance" model, where j will be
        # found by solving an algebraic equation.
        # (In the "distributed SEI resistance" model, we have already defined j)
        # For MSMR model we calculate the total current density by summing the current
        # densities from each reaction
        if domain_options["intercalation kinetics"] == "MSMR":
            j = 0
            for i in range(N):
                j0_j = self._get_exchange_current_density_by_reaction(variables, i)
                j_j = self._get_kinetics_by_reaction(j0_j, ne, eta_r, T, u, i)
                variables.update(self._get_standard_icd_by_reaction_variables(j_j, i))
                j += j_j
        else:
            j = self._get_kinetics(j0, ne, eta_r, T, u)

        if j.domain == [f"{domain} particle size"] or j.domain == [
            f"{domain} {phase_name}particle size"
        ]:
            # If j depends on particle size, get size-dependent "distribution"
            # variables first
            variables.update(
                self._get_standard_size_distribution_interfacial_current_variables(j)
            )
            variables.update(
                self._get_standard_size_distribution_exchange_current_variables(j0)
            )
            variables.update(
                self._get_standard_size_distribution_overpotential_variables(eta_r)
            )

        variables.update(self._get_standard_interfacial_current_variables(j))

        variables.update(
            self._get_standard_total_interfacial_current_variables(j_tot_av, a_j_tot_av)
        )
        variables.update(self._get_standard_exchange_current_variables(j0))
        variables.update(self._get_standard_overpotential_variables(eta_r))

        variables.update(
            self._get_standard_volumetric_current_density_variables(variables)
        )

        if self.reaction in [
            "lithium-ion main",
            "lithium metal plating",
            "lead-acid main",
        ]:
            variables.update(
                self._get_standard_sei_film_overpotential_variables(eta_sei)
            )

        return variables

    def set_algebraic(self, variables):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        if (
            self.options["total interfacial current density as a state"] == "true"
            and "main" in self.reaction
        ):
            j_tot_var = variables[
                f"Total {domain} electrode {phase_name}"
                "interfacial current density variable [A.m-2]"
            ]

            # Override print_name
            j_tot_var.print_name = "j_tot"

            a_j_tot = variables[
                f"Sum of {domain} electrode {phase_name}"
                "volumetric interfacial current densities [A.m-3]"
            ]
            a = variables[
                f"{Domain} electrode {phase_name}surface area to volume ratio [m-1]"
            ]

            # Algebraic equation to set the variable j_tot_var
            # equal to the sum of currents j_tot = a_j_tot / a
            self.algebraic[j_tot_var] = j_tot_var - a_j_tot / a

    def set_initial_conditions(self, variables):
        domain = self.domain
        phase_name = self.phase_name

        if (
            self.options["total interfacial current density as a state"] == "true"
            and "main" in self.reaction
        ):
            j_tot_var = variables[
                f"Total {domain} electrode {phase_name}"
                "interfacial current density variable [A.m-2]"
            ]
            # Set initial guess to zero
            self.initial_conditions[j_tot_var] = pybamm.Scalar(0)



#
# Class for no SEI
#
import pybamm

from .base_sei import BaseModel


class NoSEI(BaseModel):
    """
    Class for no SEI.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    options : dict
        A dictionary of options to be passed to the model.
    phase : str, optional
        Phase of the particle (default is "primary")
    cracks : bool, optional
        Whether this is a submodel for standard SEI or SEI on cracks
    """

    def __init__(self, param, domain, options, phase="primary", cracks=False):
        super().__init__(param, domain, options=options, phase=phase, cracks=cracks)
        if self.options.electrode_types[domain] == "planar":
            self.reaction_loc = "interface"
        else:
            self.reaction_loc = "full electrode"

    def get_fundamental_variables(self):
        domain = self.domain.lower()
        if self.reaction_loc == "interface":
            zero = pybamm.PrimaryBroadcast(pybamm.Scalar(0), "current collector")
        elif self.size_distribution:
            zero = pybamm.FullBroadcast(
                pybamm.Scalar(0),
                f"{domain} {self.phase_name}particle size",
                {"secondary": f"{domain} electrode", "tertiary": "current collector"},
            )
        else:
            zero = pybamm.FullBroadcast(
                pybamm.Scalar(0), f"{domain} electrode", "current collector"
            )
        variables = self._get_standard_concentration_variables(zero)

        if self.size_distribution:
            variables.update(self._get_standard_reaction_distribution_variables(zero))
            variables.update(
                self._get_standard_size_distribution_interfacial_current_variables(zero)
            )
        else:
            variables.update(self._get_standard_reaction_variables(zero))
        variables.update(self._get_standard_interfacial_current_variables(zero))
        return variables



import warnings

import pybamm

from . import BaseHysteresisOpenCircuitPotential

# Only throw the hysteresis decay rate warning once
warnings.filterwarnings(
    "once",
    message="The definition of the hysteresis decay rate parameter has changed*",
)


class OneStateHysteresisOpenCircuitPotential(BaseHysteresisOpenCircuitPotential):
    """
    Class for one-state hysteresis model based on the implementation in :footcite:t:`Axen2022`.
    The hysteresis state variable `h` is governed by an ODE which depends on the local surface
    stoichiometry and volumetric interfacial current density.

    Parameters
    ----------
    param : parameter class
        The parameters to use for this submodel
    domain : str
        The domain to implement the model, either: 'Negative' or 'Positive'.
    reaction : str
        The name of the reaction being implemented
    options: dict
        A dictionary of options to be passed to the model. See
        :class:`pybamm.BaseBatteryModel`
    phase : str, optional
        Phase of the particle (default is "primary")
    x_average : bool
        Whether the particle concentration is averaged over the x-direction. Default is False.
    """

    def __init__(
        self, param, domain, reaction, options, phase="primary", x_average=False
    ):
        warnings.warn(
            "The definition of the hysteresis decay rate parameter has changed in "
            "PyBaMM v25.10. Please see the CHANGELOG for more details.",
            stacklevel=2,
        )
        super().__init__(
            param, domain, reaction, options=options, phase=phase, x_average=x_average
        )
        pybamm.citations.register("Axen2022")

    def get_fundamental_variables(self):
        return self._get_hysteresis_state_variables()

    def get_coupled_variables(self, variables):
        return self._get_coupled_variables(variables)

    def set_rhs(self, variables):
        domain, Domain = self.domain_Domain
        phase_name = self.phase_name

        c_max = self.phase_param.c_max
        eps = self.phase_param.epsilon_s
        if self.x_average is False:
            i_vol = variables[
                f"{Domain} electrode {phase_name}volumetric interfacial current density [A.m-3]"
            ]
            sto_surf = variables[f"{Domain} {phase_name}particle surface stoichiometry"]
            T = variables[f"{Domain} electrode temperature [K]"]
            h = variables[f"{Domain} electrode {phase_name}hysteresis state"]
        else:
            i_vol = variables[
                f"X-averaged {domain} electrode {phase_name}volumetric interfacial current density [A.m-3]"
            ]
            sto_surf = variables[
                f"X-averaged {domain} {phase_name}particle surface stoichiometry"
            ]
            T = variables[f"X-averaged {domain} electrode temperature [K]"]
            h = variables[f"X-averaged {domain} electrode {phase_name}hysteresis state"]
            eps = pybamm.x_average(eps)
        gamma_lith = self.phase_param.hysteresis_decay(sto_surf, T, "lithiation")
        gamma_delith = self.phase_param.hysteresis_decay(sto_surf, T, "delithiation")

        i_vol_sign = pybamm.sign(i_vol)
        signed_h = (1 - i_vol_sign * h) / 2
        gamma = (
            0.5 * (1 - i_vol_sign) * gamma_lith + 0.5 * (1 + i_vol_sign) * gamma_delith
        )

        dhdt = gamma * i_vol / (self.param.F * c_max * eps) * signed_h

        self.rhs[h] = dhdt
